# Copyright 2011 Joe Wreschnig, Christoph Reiter
#      2013-2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys
import bz2
import itertools
from functools import reduce
from http.client import HTTPException
from os.path import splitext
from threading import Thread
from collections.abc import Collection, Callable, Iterable
from urllib.request import urlopen

import re
from gi.repository import Gtk, GLib, Pango
from senf import text2fsn

from quodlibet.util.dprint import print_d, print_e

import quodlibet
from quodlibet import _
from quodlibet import qltk
from quodlibet import util
from quodlibet import config

from quodlibet.browsers import Browser
from quodlibet.formats.remote import RemoteFile
from quodlibet.formats._audio import TAG_TO_SORT, MIGRATE, AudioFile
from quodlibet.library import SongLibrary
from quodlibet.query import Query
from quodlibet.qltk.getstring import GetStringDialog
from quodlibet.qltk.songsmenu import SongsMenu
from quodlibet.qltk.notif import Task
from quodlibet.qltk import Icons, ErrorMessage, WarningMessage
from quodlibet.util import copool, connect_destroy, sanitize_tags, connect_obj
from quodlibet.util.i18n import numeric_phrase
from quodlibet.util.path import uri_is_valid
from quodlibet.util.string import decode, encode
from quodlibet.util import print_w
from quodlibet.qltk.views import AllTreeView
from quodlibet.qltk.searchbar import SearchBarBox
from quodlibet.qltk.completion import LibraryTagCompletion
from quodlibet.qltk.x import MenuItem, Align, ScrolledWindow, Button

STATION_LIST_URL = "https://quodlibet.github.io/radio/radiolist.bz2"
STATIONS_FAV = os.path.join(quodlibet.get_user_dir(), "stations")
STATIONS_ALL = os.path.join(quodlibet.get_user_dir(), "stations_all")

# TODO: - Ranking: reduce duplicate stations (max 3 URLs per station)
#                  prefer stations that match a genre?

# Migration path for pickle
sys.modules["browsers.iradio"] = sys.modules[__name__]


class IRadioError(Exception):
    pass


class IRFile(RemoteFile):
    multisong = True
    can_add = False

    format = "Radio Station"

    __CAN_CHANGE = "title artist grouping".split()

    def __get(self, base_call, key, *args, **kwargs):
        if key == "title" and "title" not in self and "organization" in self:
            return base_call("organization", *args, **kwargs)

        # split title by " - " if no artist tag is present and
        # this is not the main song: common format for shoutcast stations
        if (
            not self.multisong
            and key in ("title", "artist")
            and "title" in self
            and "artist" not in self
        ):
            title = base_call("title").split(" - ", 1)
            if len(title) > 1:
                return (key == "title" and title[-1]) or title[0]

        if (
            key in ("artist", TAG_TO_SORT["artist"])
            and not base_call(key, *args)
            and "website" in self
        ):
            return base_call("website", *args)

        if key == "~format" and "audio-codec" in self:
            codec = base_call("audio-codec", *args, **kwargs)
            return f"{self.format} ({codec})"
        return base_call(key, *args, **kwargs)

    def __call__(self, key, *args, **kwargs):
        base_call = super().__call__
        return self.__get(base_call, key, *args, **kwargs)

    def get(self, key, *args, **kwargs):
        base_call = super().get
        return self.__get(base_call, key, *args, **kwargs)

    def write(self):
        pass

    def to_dump(self):
        # dump without title
        title = None
        if "title" in self:
            title = self["title"]
            del self["title"]
        dump = super().to_dump()
        if title is not None:
            self["title"] = title

        # add all generated tags
        lines = dump.splitlines()
        for tag in ["title", "artist", "~format"]:
            value = self.get(tag)
            if value is not None:
                lines.append(encode(tag) + b"=" + encode(value))
        return b"\n".join(lines)

    @property
    def lyric_filename(self) -> str | None:
        return None

    def can_change(self, k=None):
        if self.streamsong:
            if k is None:
                return []
            return False
        if k is None:
            return self.__CAN_CHANGE
        return k in self.__CAN_CHANGE


def parse_pls(file) -> Collection[IRFile]:
    data = {}

    lines = file.read().decode("utf-8", "replace").splitlines()

    if not lines or "[playlist]" not in lines.pop(0):
        return []

    for line in lines:
        try:
            head, val = line.strip().split("=", 1)
        except (TypeError, ValueError):
            continue
        else:
            head = head.lower()
            if head.startswith("length") and val == "-1":
                continue
            data[head] = val

    count = 1
    files = []
    warnings = []
    while True:
        if "file%d" % count in data:
            filename = text2fsn(data["file%d" % count])
            if filename.lower()[-4:] in [".pls", ".m3u", "m3u8"]:
                warnings.append(filename)
            else:
                irf = IRFile(filename)
                for key in ["title", "genre", "artist"]:
                    try:
                        irf[key] = data["%s%d" % (key, count)]
                    except KeyError:
                        pass
                try:
                    irf["~#length"] = int(data["length%d" % count])
                except (KeyError, TypeError, ValueError):
                    pass
                files.append(irf)
        else:
            break
        count += 1

    if warnings:
        raise IRadioError(
            _(
                "Station lists can only contain locations of stations, "
                "not other station lists or playlists. The following locations "
                "cannot be loaded:\n%s"
            )
            % "\n  ".join(map(util.escape, warnings))
        )
    return files


def parse_m3u(fileobj) -> Collection[IRFile]:
    files = []
    pending_title = None
    lines = fileobj.read().decode("utf-8", "replace").splitlines()
    for line in lines:
        line = line.strip()
        if line.startswith("#EXTINF:"):
            try:
                pending_title = line.split(",", 1)[1]
            except IndexError:
                pending_title = None
        elif line.startswith("http"):
            irf = IRFile(text2fsn(line))
            if pending_title:
                irf["title"] = pending_title
                pending_title = None
            files.append(irf)
    return files


def _get_stations_from(
    uri: str, on_done: Callable[[Iterable[IRFile], str], None]
) -> None:
    """Fetches the URI content and extracts IRFiles
    Called from thread - so no direct GTK+ interaction
    :param uri: URI of station
    :param on_done: a callback taking files when done (or none if errored)
    """

    with Task(_("Internet Radio"), _("Add stations")) as task:
        irfs: Collection[IRFile] = []
        GLib.idle_add(task.pulse)
        if (
            uri.lower().endswith(".pls")
            or uri.lower().endswith(".m3u")
            or uri.lower().endswith(".m3u8")
        ):
            if not re.match("^([^/:]+)://", uri):
                # Assume HTTP if no protocol given. See #2731
                uri = "http://" + uri
                print_d(f"Assuming http: {uri}")

            # Error handling outside
            sock = None
            GLib.idle_add(task.pulse)
            _fn, ext = splitext(uri.lower())
            try:
                sock = urlopen(uri, timeout=6)
                if ext == ".pls":
                    irfs = parse_pls(sock)
                elif ext in (".m3u", ".m3u8"):
                    irfs = parse_m3u(sock)
                GLib.idle_add(task.pulse)
            except OSError as e:
                print_e(f"Couldn't download from {uri} ({e})")
            finally:
                if sock:
                    sock.close()
        else:
            try:
                irfs = [IRFile(uri)]
            except ValueError as e:
                print_e(f"Can't add URI {uri}", e)
    on_done(irfs, uri)


def download_taglist(url, callback, cofuncid, step=1024 * 10):
    """Generator for loading the bz2 compressed tag list.

    Calls callback with the decompressed data or None in case of
    an error."""

    with Task(_("Internet Radio"), _("Downloading station list")) as task:
        if cofuncid:
            task.copool(cofuncid)

        try:
            response = urlopen(url)
        except (OSError, HTTPException) as e:
            print_e(f"Failed fetching from {url}", e)
            GLib.idle_add(callback, None)
            return
        try:
            size = int(response.info().get("content-length", 0))
        except ValueError:
            size = 0

        decomp = bz2.BZ2Decompressor()

        data = b""
        temp = b""
        read = 0
        while temp or not data:
            read += len(temp)

            if size:
                task.update(float(read) / size)
            else:
                task.pulse()
            yield True

            try:
                data += decomp.decompress(temp)
                temp = response.read(step)
            except (OSError, EOFError):
                data = None
                break
        response.close()

        yield True

        stations = None
        if data:
            stations = parse_taglist(data)
        print_d(f"Got {len(stations or [])} station(s)")
        GLib.idle_add(callback, stations)


def parse_taglist(data):
    """Parses a dump file like list of tags and returns a list of IRFiles

    uri=http://...
    tag=value1
    tag2=value
    tag=value2
    uri=http://...
    ...

    """

    stations = []
    station = None

    for l in data.split(b"\n"):
        if not l:
            continue
        key = l.split(b"=")[0]
        value = l.split(b"=", 1)[1]
        key = decode(key)
        value = decode(value)
        if key == "uri":
            if station:
                stations.append(station)
            station = IRFile(value)
            continue

        san = list(sanitize_tags({key: value}, stream=True).items())
        if not san:
            continue

        key, value = san[0]
        if key == "~listenerpeak":
            key = "~#listenerpeak"
            value = int(value)

        if not station:
            continue

        if isinstance(value, str):
            if value not in station.list(key):
                station.add(key, value)
        else:
            station[key] = value

    if station:
        stations.append(station)

    return stations


class AddNewStation(GetStringDialog):
    def __init__(self, parent):
        super().__init__(
            parent,
            _("New Station"),
            _("Enter the location of an Internet radio station:"),
            button_label=_("_Add"),
            button_icon=Icons.LIST_ADD,
        )

    def _verify_clipboard(self, text):
        # try to extract a URI from the clipboard
        for line in text.splitlines():
            line = line.strip()

            if uri_is_valid(line):
                return line
        return None


class GenreFilter:
    STAR = ["genre", "organization"]

    # This probably needs improvements
    GENRES = {
        "electronic": (
            _("Electronic"),
            "|(electr,house,techno,trance,/trip.?hop/,&(drum,n,bass),chill,"
            "dnb,minimal,/down(beat|tempo)/,&(dub,step))",
        ),
        "rap": (_("Hip Hop / Rap"), "|(&(hip,hop),rap)"),
        "oldies": (_("Oldies"), r"|(/[2-9]0\S?s/,oldies)"),
        "r&b": (_("R&B"), r"/r(\&|n)b/"),
        "japanese": (_("Japanese"), "|(anime,jpop,japan,jrock)"),
        "indian": (_("Indian"), "|(bollywood,hindi,indian,bhangra)"),
        "religious": (
            _("Religious"),
            "|(religious,christian,bible,gospel,spiritual,islam)",
        ),
        "charts": (_("Charts"), "|(charts,hits,top)"),
        "turkish": (_("Turkish"), "|(turkish,turkce)"),
        "reggae": (_("Reggae / Dancehall"), r"|(/reggae([^\w]|$)/,dancehall)"),
        "latin": (_("Latin"), "|(latin,salsa)"),
        "college": (_("College Radio"), "|(college,campus)"),
        "talk_news": (_("Talk / News"), "|(news,talk)"),
        "ambient": (_("Ambient"), "|(ambient,easy)"),
        "jazz": (_("Jazz"), "|(jazz,swing)"),
        "classical": (_("Classical"), "classic"),
        "pop": (_("Pop"), None),
        "alternative": (_("Alternative"), None),
        "metal": (_("Metal"), None),
        "country": (_("Country"), None),
        "news": (_("News"), None),
        "schlager": (_("Schlager"), None),
        "funk": (_("Funk"), None),
        "indie": (_("Indie"), None),
        "blues": (_("Blues"), None),
        "soul": (_("Soul"), None),
        "lounge": (_("Lounge"), None),
        "punk": (_("Punk"), None),
        "reggaeton": (_("Reggaeton"), None),
        "slavic": (_("Slavic"), "|(narodna,albanian,manele,shqip,kosova)"),
        "greek": (_("Greek"), None),
        "gothic": (_("Gothic"), None),
        "rock": (_("Rock"), None),
    }

    # parsing all above takes 350ms on an atom, so only generate when needed
    __CACHE: dict[str, Query] = {}

    def keys(self):
        return self.GENRES.keys()

    def query(self, key):
        if key not in self.__CACHE:
            text, filter_ = self.GENRES[key]
            if filter_ is None:
                filter_ = key
            self.__CACHE[key] = Query(filter_, star=self.STAR)
        return self.__CACHE[key]

    def text(self, key):
        return self.GENRES[key][0]


class CloseButton(Gtk.Button):
    """Reimplementation of 3.10 close button for InfoBar."""

    def __init__(self):
        image = Gtk.Image(
            visible=True, can_focus=False, icon_name="window-close-symbolic"
        )

        super().__init__(
            visible=False,
            can_focus=True,
            image=image,
            relief=Gtk.ReliefStyle.NONE,
            valign=Gtk.Align.CENTER,
        )

        ctx = self.get_style_context()
        ctx.add_class("raised")
        ctx.add_class("close")


class QuestionBar(Gtk.InfoBar):
    """A widget which suggest to download the radio list if
    no radio stations are present.

    Connect to Gtk.InfoBar::response and check for RESPONSE_LOAD
    as response id.
    """

    RESPONSE_LOAD = 1

    def __init__(self):
        super().__init__()
        self.connect("response", self.__response)
        self.set_message_type(Gtk.MessageType.QUESTION)

        label = Gtk.Label(
            label=_("Would you like to load a list of popular radio stations?")
        )
        label.set_line_wrap(True)
        label.show()
        content = self.get_content_area()
        content.add(label)

        self.add_button(_("_Load Stations"), self.RESPONSE_LOAD)
        self.set_show_close_button(True)

    def __response(self, bar, response_id):
        if response_id == Gtk.ResponseType.CLOSE:
            bar.hide()


class InternetRadio(Browser, util.InstanceTracker):
    __stations = None
    __fav_stations = None
    __librarian = None

    __filter = None

    name = _("Internet Radio")
    accelerated_name = _("_Internet Radio")
    keys = ["InternetRadio"]
    priority = 16
    uses_main_library = False
    headers = "title artist ~people grouping genre website ~format channel-mode".split()

    TYPE, ICON_NAME, KEY, NAME = range(4)
    TYPE_FILTER, TYPE_ALL, TYPE_FAV, TYPE_SEP, TYPE_NOCAT = range(5)
    STAR = ["artist", "title", "website", "genre", "comment"]

    @classmethod
    def _init(cls, library):
        cls.__librarian = library.librarian

        cls.__stations = SongLibrary("iradio-remote")
        cls.__stations.load(STATIONS_ALL)

        cls.__fav_stations = SongLibrary("iradio")
        cls.__fav_stations.load(STATIONS_FAV)

        cls.filters = GenreFilter()

    @classmethod
    def _destroy(cls):
        if cls.__stations.dirty:
            cls.__stations.save()
        cls.__stations.destroy()
        cls.__stations = None

        if cls.__fav_stations.dirty:
            cls.__fav_stations.save()
        cls.__fav_stations.destroy()
        cls.__fav_stations = None

        cls.__librarian = None

        cls.filters = None

    def finalize(self, restored):
        if not restored:
            # Select "All Stations" by default
            def sel_all(row):
                return row[self.TYPE] == self.TYPE_ALL

            self.view.select_by_func(sel_all, one=True)

    def __inhibit(self):
        self.view.get_selection().handler_block(self.__changed_sig)

    def __uninhibit(self):
        self.view.get_selection().handler_unblock(self.__changed_sig)

    def __destroy(self, *args):
        if not self.instances():
            self._destroy()

    def __init__(self, library, station_list_url: str = STATION_LIST_URL):
        super().__init__(spacing=12)
        self.station_list_url = station_list_url
        self.set_orientation(Gtk.Orientation.VERTICAL)

        if not self.instances():
            self._init(library)
        self._register_instance()

        self.connect("destroy", self.__destroy)

        completion = LibraryTagCompletion(self.__stations)
        self.accelerators = Gtk.AccelGroup()
        self.__searchbar = search = SearchBarBox(
            completion=completion, accel_group=self.accelerators
        )
        search.connect("query-changed", self.__filter_changed)

        def focus(widget, *args):
            qltk.get_top_parent(widget).songlist.grab_focus()

        search.connect("focus-out", focus)

        # treeview
        scrolled_window = ScrolledWindow()
        scrolled_window.show()
        scrolled_window.set_shadow_type(Gtk.ShadowType.IN)
        self.view = view = AllTreeView()
        view.show()
        view.set_headers_visible(False)
        scrolled_window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.add(view)
        model = Gtk.ListStore(int, str, str, str)

        model.append(row=[self.TYPE_ALL, Icons.FOLDER, "__all", _("All Stations")])
        model.append(row=[self.TYPE_SEP, Icons.FOLDER, "", ""])
        # Translators: Favorite radio stations
        model.append(row=[self.TYPE_FAV, Icons.FOLDER, "__fav", _("Favorites")])
        model.append(row=[self.TYPE_SEP, Icons.FOLDER, "", ""])

        filters = self.filters
        for text, k in sorted([(filters.text(k), k) for k in filters.keys()]):
            model.append(row=[self.TYPE_FILTER, Icons.EDIT_FIND, k, text])

        model.append(row=[self.TYPE_NOCAT, Icons.FOLDER, "nocat", _("No Category")])

        def separator(model, iter, data):
            return model[iter][self.TYPE] == self.TYPE_SEP

        view.set_row_separator_func(separator, None)

        def search_func(model, column, key, iter, data):
            return key.lower() not in model[iter][column].lower()

        view.set_search_column(self.NAME)
        view.set_search_equal_func(search_func, None)

        column = Gtk.TreeViewColumn("genres")
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)

        renderpb = Gtk.CellRendererPixbuf()
        renderpb.props.xpad = 3
        column.pack_start(renderpb, False)
        column.add_attribute(renderpb, "icon-name", self.ICON_NAME)

        render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)
        view.append_column(column)
        column.pack_start(render, True)
        column.add_attribute(render, "text", self.NAME)

        view.set_model(model)

        # selection
        selection = view.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.__changed_sig = connect_destroy(
            selection, "changed", util.DeferredSignal(lambda x: self.activate())
        )

        box = Gtk.HBox(spacing=6)
        box.pack_start(search, True, True, 0)
        self._searchbox = Align(box, left=0, right=6, top=0)
        self._searchbox.show_all()

        def qbar_response(infobar, response_id):
            if response_id == infobar.RESPONSE_LOAD:
                self.__update()

        self.qbar = QuestionBar()
        self.qbar.connect("response", qbar_response)
        if not self.has_stations:
            self.qbar.show()

        pane = qltk.ConfigRHPaned("browsers", "internetradio_pos", 0.4)
        vb = Gtk.VBox(spacing=0)
        vb.pack_start(scrolled_window, True, True, 0)
        fb = Gtk.FlowBox()
        fb.set_column_spacing(3)
        fb.set_homogeneous(True)
        new_station = Button(_("_Add Station…"), Icons.LIST_ADD)
        new_station.connect("clicked", self.__add)
        self._update_button = Button(_("_Update Stations"), Icons.VIEW_REFRESH)
        self._update_button.connect("clicked", self.__update)
        fb.insert(new_station, 1)
        fb.insert(self._update_button, 2)
        vb.pack_end(Align(fb, left=3), False, False, 3)
        pane.pack1(vb, resize=False, shrink=False)
        pane.show_all()

        songbox = Gtk.VBox(spacing=6)
        songbox.pack_start(self._searchbox, False, True, 0)
        self._songpane_container = Gtk.VBox()
        songbox.pack_start(self._songpane_container, True, True, 0)
        songbox.pack_start(self.qbar, False, True, 0)
        songbox.show_all()
        pane.pack2(songbox, resize=True, shrink=False)
        self.pack_start(pane, True, True, 0)
        self.show()

    @property
    def has_stations(self) -> bool:
        return bool(len(self.__stations or []) + len(self.__fav_stations or []))

    def pack(self, songpane):
        container = Gtk.VBox()
        container.add(self)
        self._songpane_container.add(songpane)
        return container

    def unpack(self, container, songpane):
        self._songpane_container.remove(songpane)
        container.remove(self)

    def __update(self, *args):
        self.qbar.hide()
        copool.add(
            download_taglist,
            self.station_list_url,
            self.__update_done,
            cofuncid="radio-load",
            funcid="radio-load",
        )

    def __update_done(self, stations):
        if not stations:
            print_w("Loading remote station list failed.")
            return

        # filter stations based on quality, listenercount
        def filter_stations(station):
            peak = station.get("~#listenerpeak", 0)
            if peak < 10:
                return False
            aac = "AAC" in station("~format")
            bitrate = station("~#bitrate", 50)
            if (aac and bitrate < 40) or (not aac and bitrate < 60):
                return False
            return True

        stations = filter(filter_stations, stations)

        # group them based on the title
        groups = {}
        for s in stations:
            key = s("~title~artist")
            groups.setdefault(key, []).append(s)

        # keep at most 2 URLs for each group
        stations = []
        for sub in groups.values():
            sub.sort(key=lambda s: s.get("~#listenerpeak", 0), reverse=True)
            stations.extend(sub[:2])

        # only keep the ones in at least one category
        all_ = [self.filters.query(k) for k in self.filters.keys()]
        assert all_
        anycat_filter = reduce(lambda x, y: x | y, all_)
        stations = list(filter(anycat_filter.search, stations))

        # remove listenerpeak
        for s in stations:
            s.pop("~#listenerpeak", None)

        # update the libraries
        stations = {s.key: s for s in stations}
        # don't add ones that are in the fav list
        for fav in self.__fav_stations.keys():
            stations.pop(fav, None)

        # separate
        o, n = set(self.__stations.keys()), set(stations)
        to_add, to_change, to_remove = n - o, o & n, o - n
        del o, n

        # migrate stats
        to_change = [stations.pop(k) for k in to_change]
        for new in to_change:
            old = self.__stations[new.key]
            # clear everything except stats
            AudioFile.reload(old)
            # add new metadata except stats
            for k in (x for x in new.keys() if x not in MIGRATE):
                old[k] = new[k]

        to_add = [stations.pop(k) for k in to_add]
        to_remove = [self.__stations[k] for k in to_remove]

        self.__stations.remove(to_remove)
        self.__stations.changed(to_change)
        self.__stations.add(to_add)

    def __filter_changed(self, bar, text, restore=False):
        self.__filter = Query(text, self.STAR)

        if not restore:
            self.activate()

    def __get_selected_libraries(self):
        """Returns the libraries to search in depending on the
        filter selection"""

        selection = self.view.get_selection()
        model, rows = selection.get_selected_rows()
        types = [model[row][self.TYPE] for row in rows]
        libs = [self.__fav_stations]
        if types != [self.TYPE_FAV]:
            libs.append(self.__stations)

        return libs

    def __get_selection_filter(self):
        """Returns a filter object for the current selection or None
        if nothing should be filtered"""

        selection = self.view.get_selection()
        model, rows = selection.get_selected_rows()

        filter_ = None
        for row in rows:
            type_ = model[row][self.TYPE]
            if type_ == self.TYPE_FILTER:
                key = model[row][self.KEY]
                current_filter = self.filters.query(key)
                if current_filter:
                    if filter_:
                        filter_ |= current_filter
                    else:
                        filter_ = current_filter
            elif type_ == self.TYPE_NOCAT:
                # if notcat is selected, combine all filters, negate and merge
                all_ = [self.filters.query(k) for k in self.filters.keys()]
                nocat_filter = all_ and -reduce(lambda x, y: x | y, all_)
                if nocat_filter:
                    if filter_:
                        filter_ |= nocat_filter
                    else:
                        filter_ = nocat_filter
            elif type_ == self.TYPE_ALL:
                filter_ = None
                break

        return filter_

    def unfilter(self):
        self.filter_text("")

    def __add_fav(self, songs):
        songs = [s for s in songs if s in self.__stations]
        type(self).__librarian.move(songs, self.__stations, self.__fav_stations)

    def __remove_fav(self, songs):
        songs = [s for s in songs if s in self.__fav_stations]
        type(self).__librarian.move(songs, self.__fav_stations, self.__stations)

    def __add(self, button):
        parent = qltk.get_top_parent(self)
        uri = (AddNewStation(parent).run(clipboard=True) or "").strip()
        if uri != "":
            self.__add_stations_from(uri)

    def __add_stations(self, irfs: Collection[IRFile], uri: str) -> None:
        print_d(f"Got {len(irfs)} station(s) from {uri}")
        assert self.__fav_stations is not None
        if not irfs:
            msg = ErrorMessage(
                self,
                _("No stations found"),
                _("No Internet radio stations were found at %s.") % util.escape(uri),
            )
            msg.run()
            return

        fav_uris = {af("~uri") for af in self.__fav_stations}
        irfs = {af for af in irfs if af("~uri") not in fav_uris}
        if irfs:
            print_d(f"Adding {irfs} to favourites")
            self.__fav_stations.add(irfs)
        else:
            message = WarningMessage(
                self,
                _("Nothing to add"),
                _("All stations listed are already in your library."),
            )
            message.run()

    def __add_stations_from(self, uri: str) -> None:
        def on_done(irfs: Iterable[IRFile], uri: str):
            GLib.idle_add(self.__add_stations, irfs, uri)
            print_d("Quitting thread")

        Thread(target=_get_stations_from, args=(uri, on_done)).start()

    def menu(self, songs, library, items):
        in_fav = False
        in_all = False
        for song in songs:
            if song in self.__fav_stations:
                in_fav = True
            elif song in self.__stations:
                in_all = True
            if in_fav and in_all:
                break

        iradio_items = []
        button = MenuItem(_("Add to Favorites"), Icons.LIST_ADD)
        button.set_sensitive(in_all)
        connect_obj(button, "activate", self.__add_fav, songs)
        iradio_items.append(button)
        button = MenuItem(_("Remove from Favorites"), Icons.LIST_REMOVE)
        button.set_sensitive(in_fav)
        connect_obj(button, "activate", self.__remove_fav, songs)
        iradio_items.append(button)

        items.append(iradio_items)
        return SongsMenu(
            self.__librarian,
            songs,
            playlists=False,
            remove=True,
            queue=False,
            items=items,
        )

    def restore(self):
        text = config.gettext("browsers", "query_text")
        self.__searchbar.set_text(text)
        if Query(text).is_parsable:
            self.__filter_changed(self.__searchbar, text, restore=True)

        keys = config.get("browsers", "radio").splitlines()

        def select_func(row):
            return row[self.TYPE] != self.TYPE_SEP and row[self.KEY] in keys

        self.__inhibit()
        view = self.view
        if not view.select_by_func(select_func):
            for row in view.get_model():
                if row[self.TYPE] == self.TYPE_FAV:
                    view.set_cursor(row.path)
                    break
        self.__uninhibit()

    def __get_filter(self):
        filter_ = self.__get_selection_filter()
        text_filter = self.__filter or Query("")

        if filter_:
            filter_ &= text_filter
        else:
            filter_ = text_filter

        return filter_

    def can_filter_text(self):
        return True

    def filter_text(self, text):
        self.__searchbar.set_text(text)
        if Query(text).is_parsable:
            self.__filter_changed(self.__searchbar, text)
            self.activate()

    def get_filter_text(self):
        return self.__searchbar.get_text()

    def activate(self):
        filter_ = self.__get_filter()
        libs = self.__get_selected_libraries()
        songs = filter_.filter(itertools.chain(*libs))
        self.songs_selected(songs)

    def active_filter(self, song):
        for lib in self.__get_selected_libraries():
            if song in lib:
                break
        else:
            return False

        filter_ = self.__get_filter()

        if filter_:
            return filter_.search(song)
        return True

    def save(self):
        text = self.__searchbar.get_text()
        config.settext("browsers", "query_text", text)

        selection = self.view.get_selection()
        model, rows = selection.get_selected_rows()
        names = filter(None, [model[row][self.KEY] for row in rows])
        config.set("browsers", "radio", "\n".join(names))

    def scroll(self, song):
        # nothing we care about
        if song not in self.__stations and song not in self.__fav_stations:
            return

        path = None
        for row in self.view.get_model():
            if row[self.TYPE] == self.TYPE_FILTER:
                if self.filters.query(row[self.KEY]).search(song):
                    path = row.path
                    break
        else:
            # in case nothing matches, select all
            path = (0,)

        self.view.set_cursor(path)
        self.view.scroll_to_cell(path, use_align=True, row_align=0.5)

    def status_text(self, count: int, time: str | None = None) -> str:
        return numeric_phrase("%(count)d station", "%(count)d stations", count, "count")


from quodlibet import app

if not app.player or app.player.can_play_uri("http://"):
    browsers = [InternetRadio]
else:
    browsers = []



# Copyright 2015 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys
import tempfile
from functools import wraps

from senf import path2fsn


from .environment import is_linux


def cmp(a, b):
    return (a > b) - (a < b)


def cached_func(f):
    """Decorator which caches the return value of a function which
    doesn't take any input.
    """

    res = []

    @wraps(f)
    def wrapper():
        if not res:
            res.append(f())
        return res[0]

    return wrapper


def total_ordering(cls):
    """Adds all possible ordering methods to a class.

    Needs a working __eq__ and __lt__ and will supply the rest.
    """

    assert "__eq__" in cls.__dict__
    assert "__lt__" in cls.__dict__

    cls.__le__ = lambda self, other: self == other or self < other
    cls.__gt__ = lambda self, other: not (self == other or self < other)
    cls.__ge__ = lambda self, other: not self < other
    cls.__ne__ = lambda self, other: not self.__eq__(other)

    return cls


def hashable(cls):
    """Makes sure the class is hashable.

    Needs a working __eq__ and __hash__ and will add a __ne__.
    """

    # py2
    assert "__hash__" in cls.__dict__
    # py3
    assert cls.__dict__["__hash__"] is not None
    assert "__eq__" in cls.__dict__

    cls.__ne__ = lambda self, other: not self.__eq__(other)

    return cls


def get_module_dir(module=None):
    """Returns the absolute path of a module. If no module is given
    the one this is called from is used.
    """

    if module is None:
        file_path = sys._getframe(1).f_globals["__file__"]
    else:
        file_path = module.__file__
    file_path = path2fsn(file_path)
    return os.path.dirname(os.path.realpath(file_path))


def get_ca_file():
    """A path to a CA file or None.

    Depends whether we use certifi or the system trust store
    on the current platform.
    """

    if is_linux():
        return None

    import certifi

    return os.path.join(get_module_dir(certifi), "cacert.pem")


def NamedTemporaryFile(*args, **kwargs):  # noqa
    """Like tempfile.NamedTemporaryFile, but supports unicode paths on
    Py2+Windows
    """

    return tempfile.NamedTemporaryFile(*args, **kwargs)



# Copyright 2004-2021 Joe Wreschnig, Michael Urman, Iñigo Serna,
#                     Steven Robertson, Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk
from gi.repository import Pango

from quodlibet.qltk.msg import ConfirmationPrompt
from senf import fsn2text

from quodlibet import _, print_w, print_d, app, ngettext
from quodlibet.qltk.chooser import choose_folders, _get_chooser, _run_chooser
from quodlibet.qltk.views import RCMHintedTreeView
from quodlibet.qltk.models import ObjectStore
from quodlibet.qltk.x import MenuItem, Button
from quodlibet.qltk import Icons
from quodlibet.util.path import unexpand
from quodlibet.util.library import get_scan_dirs, set_scan_dirs
from quodlibet.util import connect_obj, copool


class ScanBox(Gtk.HBox):
    """A box for editing the Library's scan directories"""

    def __init__(self):
        super().__init__(spacing=6)

        self.model = model = ObjectStore()
        self.view = view = RCMHintedTreeView(model=model)
        view.set_fixed_height_mode(True)
        view.set_headers_visible(False)

        menu = Gtk.Menu()
        remove_item = MenuItem(_("_Remove"), Icons.LIST_REMOVE)
        menu.append(remove_item)
        menu.show_all()
        view.connect("popup-menu", self.__popup, menu)
        connect_obj(remove_item, "activate", self.__remove, view)

        sw = Gtk.ScrolledWindow()
        sw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        sw.set_shadow_type(Gtk.ShadowType.IN)
        sw.add(view)
        sw.set_size_request(-1, max(sw.size_request().height, 80))
        sw.set_tooltip_text(
            _(
                "Songs in the listed folders will be added "
                "to the library during a library refresh"
            )
        )
        render = Gtk.CellRendererText()
        render.set_property("ellipsize", Pango.EllipsizeMode.END)

        def cdf(column, cell, model, iter_, data):
            path = model.get_value(iter_)
            cell.set_property("text", fsn2text(unexpand(path)))

        column = Gtk.TreeViewColumn(None, render)
        column.set_cell_data_func(render, cdf)
        column.set_sizing(Gtk.TreeViewColumnSizing.FIXED)
        view.append_column(column)

        add = Button(_("_Add"), Icons.LIST_ADD)
        add.set_tooltip_text(_("The new directory will be scanned after adding"))
        add.connect("clicked", self.__add)
        remove = Button(_("_Remove"), Icons.LIST_REMOVE)
        remove.set_tooltip_text(
            _(
                "All songs in the selected directories "
                "will also be removed from the library"
            )
        )

        move = Button(_("_Move"), Icons.EDIT_REDO)
        move.connect("clicked", self.__move)
        move.set_tooltip_text(
            _(
                "Move a scan root (but not the files), "
                "migrating metadata for all included tracks."
            )
        )

        selection = view.get_selection()
        selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        selection.connect("changed", self.__select_changed, remove, move)
        selection.emit("changed")

        connect_obj(remove, "clicked", self.__remove, view)

        vbox = Gtk.VBox(spacing=6)
        vbox.pack_start(add, False, True, 0)
        vbox.pack_start(remove, False, True, 0)
        vbox.pack_start(move, False, True, 0)

        self.pack_start(sw, True, True, 0)
        self.pack_start(vbox, False, True, 0)

        for path in get_scan_dirs():
            model.append(row=[path])

        for child in self.get_children():
            child.show_all()

    def __popup(self, view, menu):
        return view.popup_menu(menu, 0, Gtk.get_current_event_time())

    def __select_changed(self, selection, remove_button, move_button):
        remove_button.set_sensitive(selection.count_selected_rows())
        move_button.set_sensitive(selection.count_selected_rows() == 1)

    def __save(self):
        set_scan_dirs(list(self.model.itervalues()))

    def __remove(self, view) -> None:
        selection = self.view.get_selection()
        model, paths = selection.get_selected_rows()
        gone_dirs = [model[p][0] for p in paths or []]
        total = len(gone_dirs)
        if not total:
            return
        msg = (
            _("Remove {dir!r} and all its tracks?").format(dir=gone_dirs[0])
            if total == 1
            else _("Remove {n} library paths and their tracks?").format(n=total)
        )
        title = ngettext("Remove library path?", "Remove library paths?", total)
        prompt = ConfirmationPrompt(
            self, title, msg, _("Remove"), ok_button_icon=Icons.LIST_REMOVE
        )
        if prompt.run() == ConfirmationPrompt.RESPONSE_INVOKE:
            view.remove_selection()
            self.__save()

    def __add(self, *args):
        fns = choose_folders(self, _("Select Directories"), _("_Add Folders"))
        for fn in fns:
            self.model.append(row=[fn])
        self.__save()

    def __move(self, widget):
        selection = self.view.get_selection()
        model, paths = selection.get_selected_rows()
        rows = [model[p] for p in paths or []]
        if len(rows) > 1:
            print_w("Can't do multiple moves at once")
            return
        if not rows:
            return
        base_dir = rows[0][0]
        chooser = _get_chooser(_("Select This Directory"), _("_Cancel"))
        chooser.set_title(
            _("Select Actual / New Directory for {dir!r}").format(dir=base_dir)
        )
        chooser.set_action(Gtk.FileChooserAction.SELECT_FOLDER)
        chooser.set_local_only(True)
        chooser.set_select_multiple(False)
        results = _run_chooser(self, chooser)
        if not results:
            return
        new_dir = results[0]
        desc = _(
            "This will move QL metadata:\n\n"
            "{old!r} → {new!r}\n\n"
            "The audio files themselves are not moved by this.\n"
            "Nonetheless, a backup is recommended "
            "(including the Quod Libet 'songs' file)."
        ).format(old=base_dir, new=new_dir)
        title = _("Move scan root {dir!r}?").format(dir=base_dir)
        value = ConfirmationPrompt(
            self, title=title, description=desc, ok_button_text=_("OK, move it!")
        ).run()
        if value != ConfirmationPrompt.RESPONSE_INVOKE:
            print_d("User aborted")
            return
        print_d(f"Migrate from {base_dir} -> {new_dir}")
        copool.add(app.librarian.move_root, base_dir, new_dir)
        path = paths[0]
        self.model[path] = [new_dir]
        self.model.row_changed(path, rows[0].iter)
        self.__save()



# Copyright 2013 Simonas Kazlauskas
#           2018-23 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from os import path
from gi.repository import Soup

from quodlibet import _
from quodlibet.plugins.cover import CoverSourcePlugin, cover_dir
from quodlibet.util.cover.http import HTTPDownloadMixin, escape_query_value
from quodlibet.util.path import escape_filename


class MusicBrainzCover(CoverSourcePlugin, HTTPDownloadMixin):
    PLUGIN_ID = "musicbrainz-cover"
    PLUGIN_NAME = _("MusicBrainz Cover Source")
    PLUGIN_DESC = _("Downloads covers from MusicBrainz's cover art archive.")

    _SIZES = {
        "original": "front",
        "500x500": "front-500",
        "500x500 (back)": "back-500",
        "original (back)": "back",
    }

    @classmethod
    def group_by(cls, song):
        return song.get("musicbrainz_albumid", None)

    @staticmethod
    def priority():
        # It's a pretty good source, but very slow...
        return 0.55

    @property
    def cover_path(self):
        mbid = self.mbid
        if mbid is None:
            return super().cover_path
        return path.join(cover_dir, escape_filename(mbid))

    @property
    def mbid(self):
        return self.song.get("musicbrainz_albumid", None)

    def search(self):
        # This class has hard-coded search results,
        # and relies on 404s being filtered out later
        if not self.urls:
            return self.fail("No Musicbrainz tag found")
        self.emit(
            "search-complete",
            [{"cover": url, "dimensions": dims} for dims, url in self.urls.items()],
        )
        return None

    @property
    def urls(self) -> dict[str, str]:
        if not self.mbid:
            return {}
        mbid = escape_query_value(self.mbid)
        return {
            dim: f"https://coverartarchive.org/release/{mbid}/{extra}"
            for dim, extra in self._SIZES.items()
        }

    def fetch_cover(self):
        if not self.mbid:
            return self.fail("MBID is required to fetch the cover")
        self.download(Soup.Message.new("GET", self.urls.get("original", None)))
        return None



# Copyright 2010 Steven Robertson
#           2016 Mice Pápai
#           2022 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import shelve
import time
from datetime import date
from threading import Thread
from urllib.parse import urlencode
import json

from gi.repository import Gtk, GLib

import quodlibet
from quodlibet import _, print_w, print_e
from quodlibet import config, util, qltk
from quodlibet.qltk.entry import UndoEntry
from quodlibet.qltk import Icons
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.util.urllib import urlopen

max_wait = 15

API_KEY = "f536cdadb4c2aec75ae15e2b719cb3a1"


def apicall(method, **kwargs):
    """Performs Last.fm API call."""
    real_args = {"api_key": API_KEY, "format": "json", "method": method}
    real_args.update(kwargs)
    url = "".join(["https://ws.audioscrobbler.com/2.0/?", urlencode(real_args)])
    uobj = urlopen(url)
    json_text = uobj.read().decode("utf-8")
    resp = json.loads(json_text)
    if "error" in resp:
        errmsg = f"Last.fm API error: {resp.get('message', '')}"
        print_e(errmsg)
        raise OSError(resp["error"], errmsg)
    return resp


def config_get(key, default=None):
    return config.get("plugins", f"lastfmsync_{key}", default)


class LastFMSyncCache:
    """Stores the Last.fm charts for a particular user."""

    registered = 0
    lastupdated = None

    def __init__(self, username):
        self.username = username
        self.charts = {}
        self.songs = {}

    def update_charts(self, progress=None):
        """Updates Last.fm charts for the given user. Returns True if an
        update was attempted, False otherwise.

        progress is a callback func (msg, frac) that will be called to
        update a UI. 'frac' may be None to indicate no change should be made.
        If the function returns False, this thread will stop early."""

        def prog(msg, frac):
            if progress:
                if not progress(msg, frac):
                    # this gets caught later
                    raise ValueError()

        try:
            # Last.fm updates their charts weekly; we only poll for new
            # charts if it's been more than a day since the last poll
            if not self.registered:
                resp = apicall("user.getinfo", user=self.username)
                self.registered = int(resp["user"]["registered"]["unixtime"])

            now = time.time()
            if not self.lastupdated or self.lastupdated + (24 * 60 * 60) < now:
                prog(_("Updating chart list."), 0)
                resp = apicall("user.getweeklychartlist", user=self.username)
                charts = resp["weeklychartlist"]["chart"]
                for chart in charts:
                    # Charts keys are 2-tuple (from_timestamp, to_timestamp);
                    # values are whether we still need to fetch the chart
                    fro, to = (int(chart[s]) for s in ("from", "to"))

                    # If the chart is older than the register date of the
                    # user, don't download it. (So the download doesn't start
                    # with ~2005 every time.)
                    if to < self.registered:
                        continue

                    self.charts.setdefault((fro, to), True)
                self.lastupdated = now
            elif not [v for v in self.charts.values() if v]:
                # No charts to fetch, no update scheduled.
                prog(_("Already up-to-date."), 1.0)
                return False

            new_charts = [k for k, v in self.charts.items() if v]

            for idx, (fro, to) in enumerate(sorted(new_charts)):
                chart_week = date.fromtimestamp(fro).isoformat()
                prog(
                    _("Fetching chart for week of %s.") % chart_week,
                    (idx + 1.0) / (len(new_charts) + 2.0),
                )
                args = {"user": self.username, "from": fro, "to": to}
                try:
                    resp = apicall("user.getweeklytrackchart", **args)
                except OSError as err:
                    msg = "HTTP error %d, retrying in %d seconds."
                    print_w(msg % (err.code, max_wait))
                    for i in range(max_wait, 0, -1):
                        time.sleep(1)
                        prog(msg % (err.code, i), None)
                    resp = apicall("user.getweeklytrackchart", **args)
                try:
                    tracks = resp["weeklytrackchart"]["track"]
                except KeyError:
                    tracks = []
                # Delightfully, the API JSON frontend unboxes 1-element lists.
                if isinstance(tracks, dict):
                    tracks = [tracks]
                for track in tracks:
                    self._update_stats(track, fro, to)
                self.charts[(fro, to)] = False
            prog(_("Sync complete."), 1.0)
        except ValueError:
            # this is probably from prog()
            pass
        except Exception as e:
            util.print_exc()
            prog(_("Error during sync (%s)") % e, None)
            return False

        return True

    def _update_stats(self, track, chart_fro, chart_to):
        """Updates a single track's stats. 'track' is as returned by API;
        'chart_fro' and 'chart_to' are the chart's timestamp range."""

        # we try track mbid, (artist mbid, name), (artist name, name) as keys
        keys = []
        if track["mbid"]:
            keys.append(track["mbid"])
        for artist in (track["artist"]["mbid"], track["artist"]["#text"]):
            if artist:
                keys.append((artist.lower(), track["name"].lower()))

        stats = list(filter(None, map(self.songs.get, keys)))
        if stats:
            # Not sure if last.fm ever changes their tag values, but this
            # should map all changed values to the same object correctly
            plays = max(d.get("playcount", 0) for d in stats)
            last = max(d.get("lastplayed", 0) for d in stats)
            added = max(d.get("added", chart_to) for d in stats)
            stats = stats[0]
            stats.update({"playcount": plays, "lastplayed": last, "added": added})
        else:
            stats = {"playcount": 0, "lastplayed": 0, "added": chart_to}

        stats["playcount"] = stats["playcount"] + int(track["playcount"])
        stats["lastplayed"] = max(stats["lastplayed"], chart_fro)
        stats["added"] = min(stats["added"], chart_to)

        for key in keys:
            self.songs[key] = stats

    def update_songs(self, songs):
        """Updates each SongFile in songs from the cache."""
        for song in songs:
            keys = []
            if "musicbrainz_trackid" in song:
                keys.append(song["musicbrainz_trackid"].lower())
            if "musiscbrainz_artistid" in song:
                keys.append(
                    (
                        song["musicbrainz_artistid"].lower(),
                        song.get("title", "").lower(),
                    )
                )
            keys.append((song.get("artist", "").lower(), song.get("title", "").lower()))
            stats = list(filter(None, map(self.songs.get, keys)))
            if not stats:
                continue
            stats = stats[0]

            playcount = max(song.get("~#playcount", 0), stats["playcount"])
            if playcount != 0:
                song["~#playcount"] = playcount
            lastplayed = max(song.get("~#lastplayed", 0), stats["lastplayed"])
            if lastplayed != 0:
                song["~#lastplayed"] = lastplayed
            song["~#added"] = min(song["~#added"], stats["added"])


class LastFMSyncWindow(qltk.Dialog):
    def __init__(self, parent):
        super().__init__(_("Last.fm Sync"), parent)
        self.add_button(_("_Cancel"), Gtk.ResponseType.REJECT)
        self.add_icon_button(_("_Save"), Icons.DOCUMENT_SAVE, Gtk.ResponseType.ACCEPT)
        self.set_border_width(5)
        self.set_default_size(300, 100)

        vbox = Gtk.VBox()
        vbox.set_spacing(12)

        self.progbar = Gtk.ProgressBar()
        vbox.pack_start(self.progbar, False, True, 0)
        self.status = Gtk.Label(label="")
        vbox.pack_start(self.status, True, True, 0)
        self.get_content_area().pack_start(vbox, True, True, 0)

        self.set_response_sensitive(Gtk.ResponseType.ACCEPT, False)
        self.show_all()

    def progress(self, message, fraction):
        self.status.set_text(message)
        if fraction is not None:
            self.progbar.set_fraction(fraction)
            self.progbar.set_text("%2.1f%%" % (fraction * 100))
            if fraction == 1:
                self.set_response_sensitive(Gtk.ResponseType.ACCEPT, True)


class LastFMSync(SongsMenuPlugin):
    PLUGIN_ID = "Last.fm Sync"
    PLUGIN_NAME = _("Last.fm Sync")
    PLUGIN_DESC = _("Updates your library's statistics from your Last.fm profile.")
    PLUGIN_ICON = Icons.EMBLEM_SHARED

    CACHE_PATH = os.path.join(quodlibet.get_user_dir(), "lastfmsync.db")

    def runner(self, cache):
        changed = True
        try:
            changed = cache.update_charts(self.progress)
        except Exception as e:
            print_w(f"Couldn't update cache ({e})")
        if changed:
            self.cache_shelf[cache.username] = cache
        self.cache_shelf.close()

    def progress(self, msg, frac):
        if self.running:
            GLib.idle_add(self.dialog.progress, msg, frac)
            return True
        return False

    def plugin_songs(self, songs):
        try:
            self.cache_shelf = shelve.open(self.CACHE_PATH)
        except Exception:
            # some Python 2 DB types can't be opened in Python 3
            self.cache_shelf = shelve.open(self.CACHE_PATH, "n")

        user = config_get("username", "")
        try:
            cache = self.cache_shelf.setdefault(user, LastFMSyncCache(user))
        except Exception:
            # unpickle can fail in many ways. this is just cache, so ignore
            cache = self.cache_shelf[user] = LastFMSyncCache(user)

        self.dialog = LastFMSyncWindow(self.plugin_window)
        self.running = True
        thread = Thread(target=self.runner, args=(cache,))
        thread.daemon = True
        thread.start()
        resp = self.dialog.run()
        if resp == Gtk.ResponseType.ACCEPT:
            cache.update_songs(songs)
        self.running = False
        self.dialog.destroy()

    @classmethod
    def PluginPreferences(cls, win):
        def entry_changed(entry):
            config.set("plugins", "lastfmsync_username", entry.get_text())

        label = Gtk.Label(label=_("_Username:"), use_underline=True)
        entry = UndoEntry()
        entry.set_text(config_get("username", ""))
        entry.connect("changed", entry_changed)
        label.set_mnemonic_widget(entry)

        hbox = Gtk.HBox()
        hbox.set_spacing(6)
        hbox.pack_start(label, False, True, 0)
        hbox.pack_start(entry, True, True, 0)

        return qltk.Frame(_("Account"), child=hbox)



# Copyright 2004-2005 Joe Wreschnig, Michael Urman
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import ctypes

from quodlibet.util import load_library

from ._audio import AudioFile, translate_errors


extensions = [
    ".669",
    ".amf",
    ".ams",
    ".dsm",
    ".far",
    ".it",
    ".med",
    ".mod",
    ".mt2",
    ".mtm",
    ".okt",
    ".s3m",
    ".stm",
    ".ult",
    ".gdm",
    ".xm",
]

try:
    _modplug = load_library(["libmodplug.so.1", "libmodplug.so.0", "libmodplug-1.dll"])[
        0
    ]
except OSError:
    extensions = []
else:
    _modplug.ModPlug_GetName.argtypes = [ctypes.c_void_p]
    _modplug.ModPlug_GetName.restype = ctypes.c_char_p

    _modplug.ModPlug_Load.argtypes = [ctypes.c_void_p, ctypes.c_int]
    _modplug.ModPlug_Load.restype = ctypes.c_void_p

    _modplug.ModPlug_GetLength.argtypes = [ctypes.c_void_p]
    _modplug.ModPlug_GetLength.restype = ctypes.c_int

    _modplug.ModPlug_Unload.argtypes = [ctypes.c_void_p]
    _modplug.ModPlug_Unload.restype = None


class ModFile(AudioFile):
    format = "MOD/XM/IT"

    def __init__(self, filename):
        with translate_errors():
            data = open(filename, "rb").read()
            f = _modplug.ModPlug_Load(data, len(data))
            if not f:
                raise OSError(f"{filename!r} not a valid MOD file")
            self["~#length"] = _modplug.ModPlug_GetLength(f) // 1000
            title = _modplug.ModPlug_GetName(f) or os.path.basename(filename)
            try:
                self["title"] = title.decode("utf-8")
            except UnicodeError:
                self["title"] = title.decode("iso-8859-1")
            _modplug.ModPlug_Unload(f)

        self.sanitize(filename)

    def write(self):
        pass

    def reload(self, *args):
        artist = self.get("artist")
        super().reload(*args)
        if artist is not None:
            self.setdefault("artist", artist)

    def can_change(self, k=None):
        if k is None:
            return ["artist"]
        return k == "artist"


loader = ModFile
types = [ModFile]



# Copyright 2004-2006 Joe Wreschnig, Michael Urman, Iñigo Serna
#           2012 Christoph Reiter
#           2013 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import config
from quodlibet.plugins import PluginConfig


DEFAULT_PATTERN = (
    "<album|<album~discnumber~part~tracknumber~title~version>|"
    "<artist~title~version>>"
)


def _get_plugin_config():
    # migrate option
    if config.has_option("plugins", "trayicon_window_hide"):
        value = config.getboolean("plugins", "trayicon_window_hide")
        config.remove_option("plugins", "trayicon_window_hide")
        config.set("plugins", "icon_window_hide", value)

    pconfig = PluginConfig("icon")
    pconfig.defaults.set("window_hide", True)
    pconfig.defaults.set("tooltip", DEFAULT_PATTERN)
    pconfig.defaults.set("modifier_swap", False)

    return pconfig


pconfig = _get_plugin_config()



# Copyright 2007 Joe Wreschnig
#        2016-17 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import quodlibet.qltk.playorder


class PlayOrderPlugin(quodlibet.qltk.playorder.Order):
    """Play order plugins define alternate play orders for
    Quod Libet, of the two types: Reorder (aka Shuffle) and Repeat.
    Implementations must choose to subclass `RepeatPlugin` or `ShufflePlugin`.

    They appear, when enabled, in the combo boxes in the lower left of the
    main window, as well as in the tray icon context menu.

    If explicit "next song" button presses should be handled
    differently than reaching the end of a song, use:
        def next_implicit(self, playlist, iter): ...
        def next_explicit(self, playlist, iter): ...
        def previous_explicit(self, playlist, iter): ...
        def previous_implicit(self, playlist, iter): ...

    """

    # Note these values unset the base versions, as the plugin handler logic
    # does some auto-setting of these, based on PLUGIN_NAME, if they're None
    name: str | None = None
    display_name = None
    accelerated_name = None

    priority = 200
    """Plugins default to lower priority than built-ins"""


class RepeatPlugin(PlayOrderPlugin, quodlibet.qltk.playorder.Repeat):
    """Repeat plugins add new ways to repeat an existing,
    possibly shuffled playlist.

    Note that they must delegate to the underlying `Order` (typically a
     `Reorder`) in order for the UI to function as intended.
     As such, the only method necessary to implement from `Repeat` is
        def next(self, playlist, iter): ...

    """


class ShufflePlugin(PlayOrderPlugin, quodlibet.qltk.playorder.Reorder):
    """Shuffle plugins add new ways to reorder a given song list

    Shuffle / plugins must define at least two missing methods from `Reorder`,
    i.e.
        def next(self, playlist, iter): ...
        def previous(self, playlist, iter): ...

    """



# Synchronized Lyrics: a Quod Libet plugin for showing synchronized lyrics.
# Copyright (C) 2015 elfalem
#            2016-22 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
import functools
import os
import re
from os.path import splitext

from gi.repository import Gtk, Gdk, GLib

from quodlibet import _, util
from quodlibet import app
from quodlibet import qltk
from quodlibet.formats import AudioFile
from quodlibet.plugins import PluginConfigMixin
from quodlibet.plugins.events import EventPlugin
from quodlibet.qltk import Icons
from quodlibet.util.dprint import print_d


class SynchronizedLyrics(EventPlugin, PluginConfigMixin):
    PLUGIN_ID = "SynchronizedLyrics"
    PLUGIN_NAME = _("Synchronized Lyrics")
    PLUGIN_DESC = _(
        "Shows synchronized lyrics from an .lrc file "
        "with same name as the track (or similar)."
    )
    PLUGIN_ICON = Icons.FORMAT_JUSTIFY_FILL

    SYNC_PERIOD = 10000

    DEFAULT_BGCOLOR = "#343428282C2C"
    DEFAULT_TXTCOLOR = "#FFFFFFFFFFFF"
    DEFAULT_FONTSIZE = 25

    CFG_BGCOLOR_KEY = "backgroundColor"
    CFG_TXTCOLOR_KEY = "textColor"
    CFG_FONTSIZE_KEY = "fontSize"

    # Note the trimming of whitespace, seems "most correct" behaviour
    LINE_REGEX = re.compile(r"\s*\[([0-9]+:[0-9.]*)]\s*(.+)\s*")

    def __init__(self) -> None:
        super().__init__()
        self._lines: list[tuple[int, str]] = []
        self._timers: list[tuple[int, int]] = []
        self._start_clearing_from = 0
        self.textview = None
        self.scrolled_window = None

    @classmethod
    def PluginPreferences(cls, window):
        vb = Gtk.VBox(spacing=6)
        vb.set_border_width(6)

        t = Gtk.Table(n_rows=5, n_columns=2, homogeneous=True)
        t.set_col_spacings(6)
        t.set_row_spacings(3)

        clr_section = Gtk.Label()
        clr_section.set_markup(util.bold(_("Colors")))
        t.attach(clr_section, 0, 2, 0, 1)

        l = Gtk.Label(label=_("Text:"))
        l.set_alignment(xalign=1.0, yalign=0.5)
        t.attach(l, 0, 1, 1, 2, xoptions=Gtk.AttachOptions.FILL)

        c = Gdk.RGBA()
        c.parse(cls._get_text_color())
        b = Gtk.ColorButton(rgba=c)
        t.attach(b, 1, 2, 1, 2)
        b.connect("color-set", cls._set_text_color)

        l = Gtk.Label(label=_("Background:"))
        l.set_alignment(xalign=1.0, yalign=0.5)
        t.attach(l, 0, 1, 2, 3, xoptions=Gtk.AttachOptions.FILL)

        c = Gdk.RGBA()
        c.parse(cls._get_background_color())
        b = Gtk.ColorButton(rgba=c)
        t.attach(b, 1, 2, 2, 3)
        b.connect("color-set", cls._set_background_color)

        font_section = Gtk.Label()
        font_section.set_markup(util.bold(_("Font")))
        t.attach(font_section, 0, 2, 3, 4)

        l = Gtk.Label(label=_("Size (px):"))
        l.set_alignment(xalign=1.0, yalign=0.5)
        t.attach(l, 0, 1, 4, 5, xoptions=Gtk.AttachOptions.FILL)

        a = Gtk.Adjustment.new(cls._get_font_size(), 10, 72, 2, 3, 0)
        s = Gtk.SpinButton(adjustment=a)
        s.set_numeric(True)
        s.set_text(str(cls._get_font_size()))
        t.attach(s, 1, 2, 4, 5)
        s.connect("value-changed", cls._set_font_size)

        vb.pack_start(t, False, False, 0)
        return vb

    @classmethod
    def _get_text_color(cls):
        v = cls.config_get(cls.CFG_TXTCOLOR_KEY, cls.DEFAULT_TXTCOLOR)
        return v[:3] + v[5:7] + v[9:11]

    @classmethod
    def _get_background_color(cls):
        v = cls.config_get(cls.CFG_BGCOLOR_KEY, cls.DEFAULT_BGCOLOR)
        return v[:3] + v[5:7] + v[9:11]

    @classmethod
    def _get_font_size(cls):
        return int(cls.config_get(cls.CFG_FONTSIZE_KEY, cls.DEFAULT_FONTSIZE))

    def _set_text_color(self, button):
        self.config_set(self.CFG_TXTCOLOR_KEY, button.get_color().to_string())
        self._style_lyrics_window()

    def _set_background_color(self, button):
        self.config_set(self.CFG_BGCOLOR_KEY, button.get_color().to_string())
        self._style_lyrics_window()

    def _set_font_size(self, button):
        self.config_set(self.CFG_FONTSIZE_KEY, button.get_value_as_int())
        self._style_lyrics_window()

    def enabled(self):
        self.scrolled_window = Gtk.ScrolledWindow()
        self.scrolled_window.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC
        )
        self.scrolled_window.get_vadjustment().set_value(0)

        self.textview = Gtk.TextView()
        self.textview.set_editable(False)
        self.textview.set_cursor_visible(False)
        self.textview.set_wrap_mode(Gtk.WrapMode.WORD)
        self.textview.set_justification(Gtk.Justification.CENTER)
        self.scrolled_window.add_with_viewport(self.textview)

        vb = Gtk.HBox()
        vb.pack_start(self.scrolled_window, True, True, 6)
        vb.show_all()
        app.window.get_child().pack_start(vb, False, True, 0)
        app.window.get_child().reorder_child(vb, 2)

        self._style_lyrics_window()
        self.scrolled_window.show()

        self._sync_timer = GLib.timeout_add(self.SYNC_PERIOD, self._sync)
        self._build_data(app.player.song)
        self._timer_control()

    def disabled(self):
        self._clear_timers()
        GLib.source_remove(self._sync_timer)
        self.textview.destroy()
        self.textview = None
        self.scrolled_window.destroy()
        self.scrolled_window = None

    def _style_lyrics_window(self):
        if self.scrolled_window is None:
            return
        self.scrolled_window.set_size_request(-1, 1.5 * self._get_font_size())
        qltk.add_css(
            self.textview,
            f"""
            * {{
                background-color: {self._get_background_color()};
                color: {self._get_text_color()};
                font-size: {self._get_font_size()}px;
                padding: 0.25rem;
                border-radius: 6px;
            }}
        """,
        )

    def _cur_position(self):
        return app.player.get_position()

    @functools.lru_cache()  # noqa
    def _build_data(self, song: AudioFile | None) -> list[tuple[int, str]]:
        if self.textview:
            self.textview.get_buffer().set_text("")
        if song:
            # check in same location as track
            track_name = splitext(song("~basename") or "")[0]
            dir_ = song("~dirname")
            print_d(f"Looking for .lrc files in {dir_}")
            for filename in [
                f"{s}.lrc"
                for s in {
                    track_name,
                    track_name.lower(),
                    track_name.upper(),
                    song("~artist~title"),
                    song("~artist~tracknumber~title"),
                    song("~tracknumber~title"),
                }
            ]:
                print_d(f"Looking for {filename!r}")
                try:
                    with open(os.path.join(dir_, filename), encoding="utf-8") as f:
                        print_d(f"Found lyrics file: {filename}")
                        contents = f.read()
                except FileNotFoundError:
                    continue
                return self._parse_lrc(contents)
            print_d(f"No lyrics found for {track_name!r}")
        return []

    def _parse_lrc(self, contents: str) -> list[tuple[int, str]]:
        data = []
        for line in contents.splitlines():
            match = self.LINE_REGEX.match(line)
            if not match:
                continue
            timing, text = match.groups()
            minutes, seconds = (float(p) for p in timing.split(":", 1))
            timestamp = int(1000 * (minutes * 60 + seconds))
            data.append((timestamp, text))
        return sorted(data)

    def _set_timers(self):
        if not self._timers:
            print_d("Setting timers")
            cur_time = self._cur_position()
            cur_idx = self._greater(self._lines, cur_time)
            if cur_idx != -1:
                while (
                    cur_idx < len(self._lines)
                    and self._lines[cur_idx][0] < cur_time + self.SYNC_PERIOD
                ):
                    timestamp = self._lines[cur_idx][0]
                    line = self._lines[cur_idx][1]
                    tid = GLib.timeout_add(timestamp - cur_time, self._show, line)
                    self._timers.append((timestamp, tid))
                    cur_idx += 1

    def _sync(self):
        if not app.player.paused:
            self._clear_timers()
            self._set_timers()
        return True

    def _timer_control(self):
        if app.player.paused:
            self._clear_timers()
        else:
            self._set_timers()
        return False

    def _clear_timers(self):
        for _ts, tid in self._timers[self._start_clearing_from :]:
            GLib.source_remove(tid)
        self._timers = []
        self._start_clearing_from = 0

    def _show(self, line) -> bool:
        if self.textview:
            self.textview.get_buffer().set_text(line)
        self._start_clearing_from += 1
        print_d(f"♪ {line.strip()} ♪")
        return False

    def plugin_on_song_started(self, song: AudioFile) -> None:
        if song:
            print_d(f"Preparing for {song.key}")
        self._clear_timers()
        self._lines = self._build_data(song)
        # delay so that current position is for current track, not previous one
        GLib.timeout_add(5, self._timer_control)

    def plugin_on_song_ended(self, song, stopped):
        self._clear_timers()

    def plugin_on_paused(self):
        self._timer_control()

    def plugin_on_unpaused(self):
        self._timer_control()

    def plugin_on_seek(self, song, msec):
        if not app.player.paused:
            self._clear_timers()
            self._set_timers()

    def _greater(self, array, probe):
        length = len(array)
        if length == 0:
            return -1
        if probe < array[0][0]:
            return 0
        if probe >= array[length - 1][0]:
            return length
        return self._search(array, probe, 0, length - 1)

    def _search(self, array, probe, lower, upper):
        if lower == upper:
            if array[lower][0] <= probe:
                return lower + 1
            return lower
        middle = int((lower + upper) / 2)
        if array[middle][0] <= probe:
            return self._search(array, probe, middle + 1, upper)
        return self._search(array, probe, lower, middle)



# Copyright 2005 Joe Wreschnig
#           2016 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from quodlibet import _
from quodlibet.plugins.songshelpers import any_song
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.util.string.splitters import split_title, split_album, split_genre
from quodlibet.qltk import Icons


def has_album_splittable(song):
    return (
        "album" in song
        and "discnumber" not in song
        and song.can_change("album")
        and song.can_change("discnumber")
    )


def has_title_splittable(song):
    return "title" in song and song.can_change("title") and song.can_change("version")


def has_genre_splittable(song):
    return "genre" in song and song.can_change("genre")


class SplitTags(SongsMenuPlugin):
    PLUGIN_ID = "Split Tags"
    PLUGIN_NAME = _("Split Tags")
    PLUGIN_DESC = _(
        "Splits the disc number from the album and the version "
        "from the title at the same time."
    )
    PLUGIN_ICON = Icons.EDIT_FIND_REPLACE

    plugin_handles = any_song(has_title_splittable)

    def plugin_song(self, song):
        if has_title_splittable(song):
            title, versions = split_title(song["title"])
            if title:
                song["title"] = title
            if versions:
                song["version"] = "\n".join(versions)

        if has_album_splittable(song):
            album, disc = split_album(song["album"])
            if album:
                song["album"] = album
            if disc:
                song["discnumber"] = disc


class SplitAlbum(SongsMenuPlugin):
    PLUGIN_ID = "Split Album"
    PLUGIN_NAME = _("Split Album")
    PLUGIN_DESC = _("Split out disc number.")
    PLUGIN_ICON = Icons.EDIT_FIND_REPLACE

    plugin_handles = any_song(has_album_splittable)

    def plugin_song(self, song):
        if has_album_splittable(song):
            album, disc = split_album(song["album"])
            if album:
                song["album"] = album
            if disc:
                song["discnumber"] = disc


class SplitGenre(SongsMenuPlugin):
    PLUGIN_ID = "Split Genre"
    PLUGIN_NAME = _("Split Genre")
    PLUGIN_DESC = _("Split a single genre tag into multiple tags.")
    PLUGIN_ICON = Icons.EDIT_FIND_REPLACE

    plugin_handles = any_song(has_genre_splittable)

    def plugin_song(self, song):
        if has_genre_splittable(song):
            genre = split_genre(song["genre"])
            if genre:
                song["genre"] = genre



# Copyright 2014-2022 Nick Boultbee
#                2022 TheMelmacian
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os

from quodlibet import _, print_w, ngettext
from quodlibet import formats
from quodlibet.qltk import Icons
from quodlibet.qltk.getstring import GetStringDialog
from quodlibet.qltk.msg import ConfirmationPrompt
from quodlibet.qltk.wlw import WaitLoadWindow
from quodlibet.util import escape
from quodlibet.util.path import uri_is_valid
from urllib.response import addinfourl
from senf import uri2fsn, fsn2text, path2fsn, bytes2fsn, text2fsn


def confirm_remove_playlist_dialog_invoke(
    parent, playlist, confirmer_cls=ConfirmationPrompt
):
    """Creates and invokes a confirmation dialog that asks the user whether or not
    to go forth with the deletion of the selected playlist.

    Confirmer needs to accept the arguments for constructing a dialog,
    have a run-method returning a response, and have a RESPONSE_INVOKE
    attribute.

    returns the result of comparing the result of run to RESPONSE_INVOKE
    """
    title = _("Are you sure you want to delete the playlist '%s'?") % escape(
        playlist.name
    )
    description = _(
        "All information about the selected playlist "
        "will be deleted and can not be restored."
    )
    ok_text = _("_Delete")
    ok_icon = Icons.EDIT_DELETE

    dialog = confirmer_cls(parent, title, description, ok_text, ok_icon)
    prompt = dialog.run()
    return prompt == confirmer_cls.RESPONSE_INVOKE


def confirm_dnd_playlist_dialog_invoke(
    parent, songs, target_playlist_name, confirmer_cls=ConfirmationPrompt
):
    """see confirm_remove_playlist_dialog_invoke above, except for
    the action of attempting to extend a playlist with a second
    dragged and dropped playlist.
    """
    title = ngettext(
        'Extend "{pl_name}" with {num} additional track?',
        'Extend "{pl_name}" with {num} additional tracks?',
        len(songs),
    ).format(pl_name=target_playlist_name, num=len(songs))

    description = ""
    ok_text = _("_Add Tracks")

    dialog = confirmer_cls(parent, title, description, ok_text)
    prompt = dialog.run()
    return prompt == confirmer_cls.RESPONSE_INVOKE


class GetPlaylistName(GetStringDialog):
    def __init__(self, parent):
        super().__init__(
            parent,
            _("New Playlist"),
            _("Enter a name for the new playlist:"),
            button_label=_("_Create"),
            button_icon=Icons.DOCUMENT_NEW,
        )


def parse_m3u(filelike, pl_name, songs_lib=None, pl_lib=None):
    filenames = []
    for line in filelike:
        line = line.strip()
        if line.startswith(b"#"):
            continue
        __attempt_add(line, filenames)
    return __create_playlist(pl_name, _dir_for(filelike), filenames, songs_lib, pl_lib)


def parse_pls(filelike, pl_name, songs_lib=None, pl_lib=None):
    filenames = []
    for line in filelike:
        line = line.strip()
        if not line.lower().startswith(b"file"):
            continue
        fn = line[line.index(b"=") + 1 :].strip()
        __attempt_add(fn, filenames)
    return __create_playlist(pl_name, _dir_for(filelike), filenames, songs_lib, pl_lib)


def __attempt_add(filename, filenames):
    try:
        filenames.append(bytes2fsn(filename, "utf-8"))
    except ValueError:
        print_w(f"Ignoring invalid filename {filename!r}")


def __create_playlist(name, source_dir, files, songs_lib, pl_lib):
    songs = []
    win = WaitLoadWindow(
        None, len(files), _("Importing playlist.\n\n%(current)d/%(total)d songs added.")
    )
    win.show()
    for _i, filename in enumerate(files):
        song = None
        if not uri_is_valid(filename):
            # Plain filename.
            song = _af_for(filename, songs_lib, source_dir)
        else:
            try:
                filename = uri2fsn(filename)
            except ValueError:
                # Who knows! Hand it off to GStreamer.
                song = formats.remote.RemoteFile(filename)
            else:
                # URI-encoded local filename.
                song = _af_for(filename, songs_lib, source_dir)

        # Only add existing (not None) files to the playlist.
        # Otherwise multiple errors are thrown when the files are accessed
        # to update the displayed track infos.
        if song is not None:
            songs.append(song)
        elif os.path.exists(filename) or os.path.exists(
            os.path.join(source_dir, filename)
        ):
            print_w(
                f"Can't add file to playlist: Unsupported file format in {filename}"
            )
        else:
            print_w(f"Can't add file to playlist: File not found: {filename}")

        if win.step():
            break
    win.destroy()
    return pl_lib.create_from_songs(songs, title=name)


def _af_for(filename, library, pl_dir):
    full_path = os.path.join(pl_dir, filename)
    filename = os.path.realpath(full_path)

    af = None
    if library:
        af = library.get_filename(filename)
    if af is None:
        af = formats.MusicFile(filename)
    return af


def _name_for(filename):
    if not filename:
        return _("New Playlist")
    name = os.path.basename(os.path.splitext(filename)[0])
    return fsn2text(path2fsn(name))


def _dir_for(filelike):
    try:
        if isinstance(filelike, addinfourl):
            # if the "filelike" was created via urlopen
            # it is wrapped in an addinfourl object
            return os.path.dirname(path2fsn(filelike.fp.name))
        return os.path.dirname(path2fsn(filelike.name))
    except AttributeError:
        # Probably a URL
        return text2fsn("")



# Copyright 2004-2005 Joe Wreschnig, Michael Urman
#           2012-2023 Nick Boultbee
#                2022 Jej@github
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# Much of this code is highly optimized, because many of the functions
# are called in tight loops. Don't change things just to make them
# more readable, unless they're also faster.

import json
import os
import re
import shutil
import time
from typing import Any, Generic, TypeVar
from collections import OrderedDict
from itertools import zip_longest

from senf import fsn2uri, fsnative, fsn2text, bytes2fsn, path2fsn

from quodlibet import _, print_d
from quodlibet import util
from quodlibet import config
from quodlibet.util.path import (
    mkdir,
    mtime,
    normalize_path,
    ismount,
    get_home_dir,
    RootPathFile,
)
from quodlibet.util.string import encode, decode, isascii
from quodlibet.util.environment import is_windows

from quodlibet.util import iso639, cached_property
from quodlibet.util import human_sort_key as human, capitalize

from quodlibet.util.string.date import format_date
from quodlibet.util.tags import TAG_ROLES, TAG_TO_SORT

from ._image import ImageContainer
from ._misc import AudioFileError, translate_errors


translate_errors  # noqa

AlbumKey = tuple[str, str, str]
"""An album key is (currently) a tuple"""

MIGRATE = {
    "~#playcount",
    "~#laststarted",
    "~#lastplayed",
    "~#added",
    "~#skipcount",
    "~#rating",
    "~bookmark",
}
"""These get migrated if a song gets reloaded"""

PEOPLE = [
    "artist",
    "albumartist",
    "author",
    "composer",
    "~performers",
    "originalartist",
    "producer",
    "lyricist",
    "arranger",
    "conductor",
]
"""Sources of the ~people tag, most important first"""

TIME_TAGS = {"~#lastplayed", "~#laststarted", "~#added", "~#mtime"}
"""Time in seconds since epoch, defaults to 0"""

HUMAN_TO_NUMERIC_TIME_TAGS = {t.replace("~#", "~"): t for t in TIME_TAGS}

DURATION_TAGS = {"~#length"}
"""Duration in seconds"""

SIZE_TAGS = {"~#filesize"}
"""Size in bytes, defaults to 0"""

NUMERIC_ZERO_DEFAULT = {"~#skipcount", "~#playcount", "~#length", "~#bitrate"}
"""Defaults to 0"""

NUMERIC_ZERO_DEFAULT.update(TIME_TAGS)
NUMERIC_ZERO_DEFAULT.update(SIZE_TAGS)

FILESYSTEM_TAGS = {"~filename", "~basename", "~dirname", "~mountpoint"}
"""Values are bytes in Linux instead of unicode"""

SORT_TO_TAG = {v: k for (k, v) in TAG_TO_SORT.items()}
"""Reverse map, so sort tags can fall back to the normal ones"""

PEOPLE_SORT = [TAG_TO_SORT.get(k, k) for k in PEOPLE]
"""Sources for ~peoplesort, most important first"""

VARIOUS_ARTISTS_VALUES = "V.A.", "various artists", "Various Artists"
"""Values for ~people representing lots of people, most important last"""


def decode_value(tag, value):
    """Returns a unicode representation of the passed value, based on
    the type and the tag it originated from.

    Not reversible.
    """

    if tag in FILESYSTEM_TAGS:
        return fsn2text(value)
    if tag[:2] == "~#":
        if isinstance(value, float):
            return f"{value:.2f}"
        return str(value)
    return str(value)


K = TypeVar("K")


class HasKey(Generic[K]):
    """Many things can be keyed"""

    key: K


class AudioFile(dict, ImageContainer, HasKey):
    """An audio file. It looks like a dict, but implements synthetic
    and tied tags via __call__ rather than __getitem__. This means
    __getitem__, get, and so on can be used for efficiency.

    If you need to sort many AudioFiles, you can use their sort_key
    attribute as a decoration.

    Keys are either ASCII str or unicode.
    Values are always unicode except if the tag is part of FILESYSTEM_TAGS,
    then the value is of the path type (str on UNIX, unicode on Windows)

    Some methods will make sure the returned values are always unicode, see
    their description.
    """

    fill_metadata = False
    """New tags received from the backend will update the song"""

    fill_length = False
    """New song duration from the backend will update the song"""

    multisong = False
    """Is a container for multiple songs
     i.e. while played new songs can start / end"""

    streamsong = False
    """Is part of a multisong"""

    can_add = True
    """Can be added to the queue, playlists"""

    is_file = True
    """Is a real (local) file"""

    format = "Unknown Audio File"
    """The underlying file format"""

    supports_rating_and_play_count_in_file = False
    """Does this format support storing ratings and play counts in the file"""

    mimes: list[str] = []
    """MIME types this class can represent"""

    @cached_property
    def _date_format(self) -> str:
        return config.gettext("settings", "datecolumn_timestamp_format")

    def __init__(self, default=(), **kwargs):
        for key, value in dict(default).items():
            self[key] = value
        for key, value in kwargs.items():
            self[key] = value

    def __song_key(self):
        return (
            self("~#disc", 0),
            self("~#track", 0),
            human(self("artistsort")),
            self.get("musicbrainz_artistid", ""),
            human(self.get("title", "")),
            self.get("~filename"),
        )

    @util.cached_property
    def album_key(self) -> AlbumKey:
        id_val: str = (
            self.get("album_grouping_key")
            or self.get("labelid")
            or self.get("musicbrainz_albumid", "")
        )  # type: ignore
        return id_val, human(self("albumsort", "")), human(self("albumartistsort", ""))

    @util.cached_property
    def sort_key(self):
        return [self.album_key, self.__song_key()]

    @staticmethod
    def sort_by_func(tag):
        """Returns a fast sort function for a specific tag (or pattern).
        Some keys are already in the sort cache, so we can use them."""

        def artist_sort(song):
            return song.sort_key[1][2]

        if callable(tag):
            return lambda song: human(tag(song))
        if tag == "artistsort":
            return artist_sort
        if tag in FILESYSTEM_TAGS:
            return lambda song: fsn2text(song(tag))
        if tag.startswith("~#") and "~" not in tag[2:]:
            return lambda song: song(tag, 0)
        return lambda song: human(song(tag))

    def __getstate__(self):
        """Don't pickle anything from __dict__"""

    def __setstate__(self, state):
        """Needed because we have defined getstate"""

    def __setitem__(self, key, value):
        # validate key
        if not isinstance(key, str):
            raise TypeError("key has to be str")

        # validate value
        if key.startswith("~#"):
            if not isinstance(value, int | float):
                raise TypeError
        elif key in FILESYSTEM_TAGS:
            if not isinstance(value, fsnative):
                value = path2fsn(value)
        else:
            value = str(value)

        dict.__setitem__(self, key, value)

        pop = self.__dict__.pop
        pop("album_key", None)
        pop("sort_key", None)

    def __delitem__(self, key):
        dict.__delitem__(self, key)

        pop = self.__dict__.pop
        pop("album_key", None)
        pop("sort_key", None)

    @property
    def key(self) -> K:  # type: ignore
        return self["~filename"]

    @property
    def mountpoint(self):
        return self["~mountpoint"]

    def __hash__(self):
        # Dicts aren't hashable by default, so we need a hash
        # function. Previously this used ~filename. That created a
        # situation when an object could end up in two buckets by
        # renaming files. So now it uses identity.
        return hash(id(self))

    def __eq__(self, other):
        # And to preserve Python hash rules, we need a strict __eq__.
        return self is other

    def __lt__(self, other):
        return self.sort_key < other.sort_key

    def __ne__(self, other):
        return self is not other

    def has_rating_and_playcount_in_file(self, email):
        """Returns True if the audio file has rating and play counts for the
        provided email. Returns False otherwise, or if storing ratings and play
        counts in the file is not supported for this audio format.
        """
        return False

    def reload(self):
        """Reload an audio file from disk. If reloading fails nothing will
        change.

        Raises:
            AudioFileError: if the file fails to load
        """

        backup = dict(self)
        fn = self["~filename"]
        saved = {}
        persisted = config.getboolean("editing", "save_to_songs")
        persisted_keys = (
            {"~#rating", "~#playcount"}
            if self.supports_rating_and_play_count_in_file and persisted
            else set()
        )
        for key in self:
            # Only migrate keys that aren't (probably) persisted to file (#3569)
            if key in MIGRATE - persisted_keys:
                saved[key] = self[key]
        self.clear()
        self["~filename"] = fn
        try:
            self.__init__(fn)
        except AudioFileError:
            self.update(backup)
            raise
        else:
            self.update(saved)

    def realkeys(self):
        """Returns a list of keys that are not internal, i.e. they don't
        have '~' in them."""

        return [s for s in self.keys() if s[:1] != "~"]

    def prefixkeys(self, prefix):
        """Returns a list of dict keys that either match prefix or start
        with prefix + ':'.
        """

        l = []
        for k in self:
            if k.startswith(prefix):
                if k == prefix or k.startswith(prefix + ":"):
                    l.append(k)
        return l

    def _prefixvalue(self, tag):
        return "\n".join(self.list_unique(sorted(self.prefixkeys(tag))))

    def iterrealitems(self):
        return ((k, v) for (k, v) in self.items() if k[:1] != "~")

    def __call__(self, key, default: Any = "", connector=" - ", joiner=", "):
        """Return the value(s) for a key, synthesizing if necessary.
        Multiple values for a key are delimited by newlines.

        A default value may be given (like `dict.get`);
        the default default is an empty unicode string
        (even if the tag is numeric).

        If a tied tag ('a~b') is requested, the `connector` keyword
        argument may be used to specify what it is tied with.
        In case the tied tag contains numeric and file path tags, the result
        will still be a unicode string.
        The `joiner` keyword specifies how multiple *values* will be joined
        within that tied tag output, e.g.
            ~people~title = "Kanye West, Jay Z - New Day"

        For details on tied tags, see the documentation for `util.tagsplit`.
        """
        real_key = key
        if key[:1] == "~":
            key = key[1:]
            if "~" in key:
                values = []
                sub_tags = util.tagsplit(real_key)
                # If it's genuinely a tied tag (not ~~people etc.), we want
                # to delimit the multi-values separately from the tying
                j = joiner if len(sub_tags) > 1 else "\n"
                for t in sub_tags:
                    vs = [decode_value(real_key, v) for v in (self.list(t))]
                    v = j.join(vs)
                    if v:
                        values.append(v)
                return connector.join(values) or default
            if key == "#track":
                try:
                    return int(self["tracknumber"].split("/")[0])
                except (ValueError, TypeError, KeyError):
                    return default
            elif key == "#disc":
                try:
                    return int(self["discnumber"].split("/")[0])
                except (ValueError, TypeError, KeyError):
                    return default
            elif key == "length":
                length = self.get("~#length")
                if length is None:
                    return default
                return util.format_time_display(length)
            elif key == "#rating":
                return dict.get(self, real_key, config.RATINGS.default)
            elif key == "rating":
                return util.format_rating(self("~#rating"))
            elif key == "people":
                return "\n".join(self.list_unique(PEOPLE)) or default
            elif key == "people:real":
                # Issue 1034: Allow removal of V.A. if others exist.
                unique = self.list_unique(PEOPLE)
                # Order is important, for (unlikely case): multiple removals
                for val in VARIOUS_ARTISTS_VALUES:
                    if len(unique) > 1 and val in unique:
                        unique.remove(val)
                return "\n".join(unique) or default
            elif key == "people:roles":
                return self._role_call("performer", PEOPLE) or default
            elif key == "peoplesort":
                return "\n".join(self.list_unique(PEOPLE_SORT)) or self(
                    "~people", default, connector
                )
            elif key == "peoplesort:roles":
                # Ignores non-sort tags if there are any sort tags (e.g. just
                # returns "B" for {artist=A, performersort=B}).
                # TODO: figure out the "correct" behavior for mixed sort tags
                return self._role_call("performersort", PEOPLE_SORT) or self(
                    "~peoplesort", default, connector
                )
            elif key in ("performers", "performer"):
                return self._prefixvalue("performer") or default
            elif key in ("performerssort", "performersort"):
                return self._prefixvalue("performersort") or self(
                    real_key[-4:], default, connector
                )
            elif key in ("performers:roles", "performer:roles"):
                return self._role_call("performer") or default
            elif key in ("performerssort:roles", "performersort:roles"):
                return self._role_call("performersort") or self(
                    real_key.replace("sort", ""), default, connector
                )
            elif key == "basename":
                return os.path.basename(self["~filename"]) or self["~filename"]
            elif key == "dirname":
                return os.path.dirname(self["~filename"]) or self["~filename"]
            elif key == "uri":
                try:
                    return self["~uri"]
                except KeyError:
                    return fsn2uri(self["~filename"])
            elif key == "format":
                return self.get("~format", str(self.format))
            elif key == "codec":
                codec = self.get("~codec")
                if codec is None:
                    return self("~format")
                return codec
            elif key == "encoding":
                encoding = "\n".join(
                    part
                    for part in [self.get("~encoding"), self.get("encodedby")]
                    if part
                )
                return encoding or default
            elif key == "language":
                codes = self.list("language")
                if not codes:
                    return default
                return "\n".join(iso639.translate(c) or c for c in codes)
            elif key == "bitrate":
                return util.format_bitrate(self("~#bitrate"))
            elif key == "#date":
                date = self.get("date")
                if date is None:
                    return default
                return util.date_key(date)
            elif key == "year":
                return util.parse_year(self.get("date", default))
            elif key == "#year":
                try:
                    return int(util.parse_year(self.get("date", default)))
                except (ValueError, TypeError, KeyError):
                    return default
            elif key == "originalyear":
                return util.parse_year(self.get("originaldate", default))
            elif key == "#originalyear":
                try:
                    return int(util.parse_year(self.get("originaldate", default)))
                except (ValueError, TypeError, KeyError):
                    return default
            elif key == "#tracks":
                try:
                    return int(self["tracknumber"].split("/")[1])
                except (ValueError, IndexError, TypeError, KeyError):
                    return default
            elif key == "#discs":
                try:
                    return int(self["discnumber"].split("/")[1])
                except (ValueError, IndexError, TypeError, KeyError):
                    return default
            elif key == "lyrics":
                # First, try the embedded lyrics.
                try:
                    return self["lyrics"]
                except KeyError:
                    pass

                try:
                    return self["unsyncedlyrics"]
                except KeyError:
                    pass

                # If there are no embedded lyrics, try to read them from
                # the external file.
                lyric_filename = self.lyric_filename
                if not lyric_filename:
                    return default
                try:
                    with open(lyric_filename, "rb") as fileobj:
                        print_d(f"Reading lyrics from {lyric_filename!r}")
                        text = fileobj.read().decode("utf-8", "replace")
                        # try to skip binary files
                        if "\0" in text:
                            return default
                        return text
                except (OSError, UnicodeDecodeError):
                    return default
            elif key == "filesize":
                return util.format_size(self("~#filesize", 0))
            elif key == "playlists":
                # TODO: avoid static dependency here... somehow
                from quodlibet import app

                lib = app.library
                if not lib:
                    return ""
                playlists = lib.playlists.playlists_featuring(self)
                return "\n".join(s.name for s in playlists) or default
            elif key.startswith("#replaygain_"):
                try:
                    val = self.get(key[1:], default)
                    return round(float(val.split(" ")[0]), 2)
                except (ValueError, TypeError, AttributeError):
                    return default
            elif real_key in HUMAN_TO_NUMERIC_TIME_TAGS:
                time_value = float(self.get(HUMAN_TO_NUMERIC_TIME_TAGS[real_key], 0))
                return format_date(time_value, self._date_format)
            elif key[:1] == "#":
                if real_key in self:
                    return self[real_key]
                if real_key in NUMERIC_ZERO_DEFAULT:
                    return 0
                try:
                    val = self[real_key[2:]]
                except KeyError:
                    return default
                try:
                    return int(val)
                except ValueError:
                    try:
                        return float(val)
                    except ValueError:
                        return default
            elif key == "json":
                # Help the testing by being deterministic with sort_keys.
                return json.dumps(self, sort_keys=True)
            else:
                return dict.get(self, real_key, default)

        elif key == "title":
            title = dict.get(self, "title")
            if title is None:
                # build a title with missing_title_template option
                unknown_track_template = _(
                    config.gettext("browsers", "missing_title_template")
                )

                from quodlibet.pattern import Pattern

                try:
                    pattern = Pattern(unknown_track_template)
                except ValueError:
                    title = decode_value("~basename", self("~basename"))
                else:
                    title = pattern % self

            return title
        elif key in SORT_TO_TAG:
            try:
                return self[key]
            except KeyError:
                key = SORT_TO_TAG[key]
        return dict.get(self, key, default)

    def _role_call(self, role_tag, sub_keys=None):
        role_tag_keys = self.prefixkeys(role_tag)

        role_map = {}
        for key in role_tag_keys:
            if key == role_tag:
                # #2986: don't add a role description for the bare tag, unless
                # this is a composite tag (e.g. only show "(Performance)" for
                # ~people:roles and not ~performer:roles).
                if sub_keys is None:
                    continue
                role = TAG_ROLES.get(role_tag, role_tag)
            else:
                role = key.split(":", 1)[-1]
            for name in self.list(key):
                role_map.setdefault(name, []).append(role)

        if sub_keys is None:
            names = self.list_unique(role_tag_keys)
        else:
            names = self.list_unique(sub_keys)
            for tag in sub_keys:
                if tag in TAG_ROLES:
                    for name in self.list(tag):
                        role_map.setdefault(name, []).append(TAG_ROLES[tag])

        descs = []
        for name in names:
            roles = role_map.get(name, [])
            if not roles:
                descs.append(name)
            else:
                roles = sorted(map(capitalize, roles))
                descs.append(f"{name} ({', '.join(roles)})")

        return "\n".join(descs)

    @property
    def lyric_filename(self) -> str | None:
        """Returns the validated, or default, lyrics filename for this
        file. User defined '[memory] lyric_rootpaths' and
        '[memory] lyric_filenames' matches take precedence"""

        from quodlibet.pattern import ArbitraryExtensionFileFromPattern

        rx_params = re.compile(r"[^\\]<[^" + re.escape(os.sep) + r"]*[^\\]>")

        def expand_pathfile(rpf):
            """Return the expanded RootPathFile"""
            expanded = []
            root = os.path.expanduser(rpf.root)
            pathfile = os.path.expanduser(rpf.pathfile)
            if rx_params.search(pathfile):
                root = ArbitraryExtensionFileFromPattern(root).format(self)
                pathfile = ArbitraryExtensionFileFromPattern(pathfile).format(self)
            rpf = RootPathFile(root, pathfile)
            expanded.append(rpf)
            if not os.path.exists(pathfile) and is_windows():
                # prioritise a special character encoded version
                #
                # most 'alien' chars are supported for 'nix fs paths, and we
                # only pass the proposed path through 'escape_filename' (which
                # apparently doesn't respect case) if we don't care about case!
                #
                # FIX: assumes 'nix build used on a case-sensitive fs, nt case
                # insensitive. clearly this is not biting anyone though (yet!)
                pathfile = os.path.sep.join([rpf.root, rpf.end_escaped])
                rpf = RootPathFile(rpf.root, pathfile)
                expanded.insert(len(expanded) - 1, rpf)
            return expanded

        def sanitise(sep, parts):
            """Return a sanitised version of a path's parts"""
            return sep.join(part.replace(os.path.sep, "")[:128] for part in parts)

        # setup defaults (user-defined take precedence)
        # root search paths
        lyric_paths = config.getstringlist("memory", "lyric_rootpaths", [])
        # ensure default paths
        lyric_paths.append(os.path.join(get_home_dir(), ".lyrics"))
        lyric_paths.append(os.path.join(os.path.dirname(self.comma("~filename"))))
        # search pathfile names
        lyric_filenames = config.getstringlist("memory", "lyric_filenames", [])
        # ensure some default pathfile names
        lyric_filenames.append(
            sanitise(
                os.sep,
                [(self.comma("lyricist") or self.comma("artist")), self.comma("title")],
            )
            + ".lyric"
        )
        lyric_filenames.append(
            sanitise(
                " - ",
                [(self.comma("lyricist") or self.comma("artist")), self.comma("title")],
            )
            + ".lyric"
        )

        # generate all potential paths (unresolved/unexpanded)
        pathfiles = OrderedDict()
        for r in lyric_paths:
            for f in lyric_filenames:
                pathfile = os.path.join(
                    r, os.path.dirname(f), fsnative(os.path.basename(f))
                )
                rpf = RootPathFile(r, pathfile)
                if pathfile not in pathfiles:
                    pathfiles[pathfile] = rpf

        # expand each raw pathfile in turn and test for existence
        match_ = None
        pathfiles_expanded = OrderedDict()
        for _pf, rpf in pathfiles.items():
            for rpf in expand_pathfile(rpf):  # resolved as late as possible  # noqa
                pathfile = rpf.pathfile
                pathfiles_expanded[pathfile] = rpf
                if os.path.exists(pathfile):
                    match_ = pathfile
                    break
            if match_:
                break

        if not match_:
            # search even harder!
            lyric_extensions = ["lyric", "lyrics", "", "txt"]

            def generate_mod_ext_paths(pathfile):
                # separate pathfile's extension (if any)
                ext = os.path.splitext(pathfile)[1][1:]
                path = pathfile[: -1 * len(ext)].strip(".") if ext else pathfile
                # skip the proposed lyric extension if it is the same as
                # the original for a given search pathfile stub - it has
                # already been tested without success!
                extra_extensions = [x for x in lyric_extensions if x != ext]

                # join valid new extensions to pathfile stub and return
                return [f"{path}.{ext}" if ext else path for ext in extra_extensions]

            # look for a match by modifying the extension for each of the
            # (now fully resolved) 'pathfiles_expanded' search items
            for pathfile in pathfiles_expanded.keys():
                # get alternatives for existence testing
                paths_mod_ext = generate_mod_ext_paths(pathfile)
                for path_ext in paths_mod_ext:
                    if os.path.exists(path_ext):
                        # persistence has paid off!
                        match_ = path_ext
                        break
                if match_:
                    break

        if not match_:
            # default
            match_ = list(pathfiles_expanded.keys())[0]

        return match_

    @property
    def has_rating(self):
        """True if the song has a rating set.

        In case this is False song('~#rating') would return the default value
        """

        return self.get("~#rating") is not None

    def remove_rating(self):
        """Removes the set rating so the default will be returned"""

        self.pop("~#rating", None)

    def comma(self, key):
        """Get all values of a tag, separated by commas. Synthetic
        tags are supported, but will be slower. All list items
        will be unicode.

        If the value is numeric, that is returned rather than a list.

        Because this function is used to format values for display,
        blank tags are removed from the results.
        """

        if "~" in key or key == "title":
            if key in FILESYSTEM_TAGS:
                v = fsn2text(self(key, fsnative()))
            else:
                v = self(key, "")
        else:
            v = self.get(key, "")

        if isinstance(v, int | float):
            return v
        return re.sub("\n+", ", ", v.strip())

    def list(self, key) -> list[Any]:
        """Get all values of a tag, as a list. Synthetic tags are supported,
        but will be slower. Numeric tags will give their one value.

        For file path keys the returned list might contain path items
        (non-unicode).

        An empty tag cannot be distinguished from a non-existent tag; both
        result in []. If a file contains multiple values of a tag, empty
        instances of the tag will not be included in the list.
        """

        if "~" in key or key == "title":
            v = self(key)
            if v == "":
                return []
            v = v.split("\n") if isinstance(v, str) else [v]
            return [x for x in v if x]
        v = self.get(key, "")
        return [x for x in v.split("\n") if x]

    def list_sort(self, key):
        """Like list but return display,sort pairs when appropriate
        and work on all tags.

        In case no sort value exists the display one is returned. If a
        display tag is empty, the tag is removed from the returned list
        and the corresponding tag in the sort tag list is ignored as
        well.

        The assumption being made here is that if the ordering between
        display and sort tags doesn't match exactly, that's a problem
        with the user's file.
        """

        display = decode_value(key, self(key))
        display = display.split("\n") if display else []
        sort = []
        if key in TAG_TO_SORT:
            sort = decode_value(TAG_TO_SORT[key], self(TAG_TO_SORT[key]))
            # it would be better to use something that doesn't fall back
            # to the key itself, but what?
            sort = sort.split("\n") if sort else []

        result = []
        for d, s in zip_longest(display, sort):
            if d:
                result.append((d, (s if s is not None and s != "" else d)))
        return result

    def list_separate(self, key):
        """For tied tags return the list union of the display,sort values
        otherwise just do list_sort
        """
        if key[:1] == "~" and "~" in key[1:]:  # tied tag
            vals = [self.list_sort(tag) for tag in util.tagsplit(key)]
            return [j for i in vals for j in i]
        return self.list_sort(key)

    def list_unique(self, keys):
        """Returns a combined value of all values in keys; duplicate values
        will be ignored.

        Returns the same as list().
        """

        l = []
        seen = set()
        for k in keys:
            for v in self.list(k):
                if v not in seen:
                    l.append(v)
                    seen.add(v)
        return l

    def as_lowercased(self):
        """Returns a new AudioFile with all keys lowercased / values merged.

        Useful for tag writing for case insensitive tagging formats like
        APEv2 or VorbisComment.
        """

        merged = AudioFile()
        text = {}
        for key, value in self.items():
            lower = key.lower()
            if key.startswith("~#"):
                merged[lower] = value
            else:
                text.setdefault(lower, []).extend(value.split("\n"))
        for key, values in text.items():
            merged[key] = "\n".join(values)
        return merged

    def exists(self):
        """Return true if the file still exists (or we can't tell)."""

        return os.path.exists(self["~filename"])

    def valid(self):
        """Return true if the file cache is up-to-date (checked via
        mtime), or we can't tell."""
        return bool(self.get("~#mtime", 0)) and self["~#mtime"] == mtime(
            self["~filename"]
        )

    def mounted(self):
        """Return true if the disk the file is on is mounted, or
        the file is not on a disk."""
        return ismount(self.get("~mountpoint", "/"))

    def can_multiple_values(self, key=None):
        """If no arguments are given, return a list of tags that can
        have multiple values, or True if 'any' tags can.
        """

        return True

    def can_change(self, k=None):
        """See if this file supports changing the given tag. This may
        be a limitation of the file type or QL's design.

        The writing code should handle all kinds of keys, so this is
        just a suggestion.

        If no arguments are given, return a list of tags that can be
        changed, or True if 'any' tags can be changed (specific tags
        should be checked before adding)."""

        if k is None:
            return True

        if not isascii(k):
            return False

        if not k or "=" in k or "~" in k:
            return False

        return True

    def is_writable(self):
        return os.access(self["~filename"], os.W_OK)

    def rename(self, newname):
        """Rename a file. Errors are not handled. This shouldn't be used
        directly; use library.rename instead."""

        if os.path.isabs(newname):
            mkdir(os.path.dirname(newname))
        else:
            newname = os.path.join(self("~dirname"), newname)

        if not os.path.exists(newname):
            shutil.move(self["~filename"], newname)
        elif normalize_path(newname, canonicalise=True) != self["~filename"]:
            raise ValueError

        self.sanitize(newname)

    def sanitize(self, filename=None):
        """Fill in metadata defaults. Find ~mountpoint, ~#mtime, ~#filesize
        and ~#added. Check for null bytes in tags.

        Does not raise.
        """

        # Replace nulls with newlines, trimming zero-length segments
        for key, val in list(self.items()):
            self[key] = val
            if isinstance(val, str) and "\0" in val:
                self[key] = "\n".join(filter(lambda s: s, val.split("\0")))
            # Remove unnecessary defaults
            if key in NUMERIC_ZERO_DEFAULT and val == 0:
                del self[key]

        if filename:
            self["~filename"] = filename
        elif "~filename" not in self:
            raise ValueError("Unknown filename!")

        assert isinstance(self["~filename"], fsnative)

        if self.is_file:
            self["~filename"] = normalize_path(self["~filename"], canonicalise=True)
            # Find mount point (terminating at "/" if necessary)
            head = self["~filename"]
            while "~mountpoint" not in self:
                head, tail = os.path.split(head)
                # Prevent infinite loop without a fully-qualified filename
                # (the unit tests use these).
                head = head or fsnative("/")
                if ismount(head):
                    self["~mountpoint"] = head
        else:
            self["~mountpoint"] = fsnative("/")

        # Fill in necessary values.
        self.setdefault("~#added", int(time.time()))

        # For efficiency, do a single stat here. See Issue 504
        try:
            stat = os.stat(self["~filename"])
            self["~#mtime"] = stat.st_mtime
            self["~#filesize"] = stat.st_size

            # Issue 342. This is a horrible approximation (due to headers) but
            # on FLACs, the most common case, this should be close enough
            if "~#bitrate" not in self:
                try:
                    # Note - kbps = bytes * 8 / seconds / 1000
                    self["~#bitrate"] = int(
                        stat.st_size / (self["~#length"] * (1000 / 8))
                    )
                except (KeyError, ZeroDivisionError):
                    pass
        except OSError:
            self["~#mtime"] = 0

    def to_dump(self):
        """A string of 'key=value' lines, similar to vorbiscomment output.

        Returns:
            bytes
        """

        def encode_key(k):
            return encode(k) if isinstance(k, str) else k

        s = []
        for k in self.keys():
            enc_key = encode_key(k)
            assert isinstance(enc_key, bytes)

            if isinstance(self[k], int):
                l = enc_key + encode("=%d" % self[k])
                s.append(l)
            elif isinstance(self[k], float):
                l = enc_key + encode(f"={self[k]:f}")
                s.append(l)
            else:
                for v2 in self.list(k):
                    if not isinstance(v2, bytes):
                        v2 = encode(v2)
                    s.append(enc_key + b"=" + v2)
        for k in NUMERIC_ZERO_DEFAULT - set(self.keys()):
            enc_key = encode_key(k)
            l = enc_key + encode("=%d" % self.get(k, 0))
            s.append(l)
        if "~#rating" not in self:
            s.append(encode("~#rating={:f}".format(self("~#rating"))))
        s.append(encode(f"~format={self.format}"))
        s.append(b"")
        return b"\n".join(s)

    def from_dump(self, text):
        """Parses the text created with to_dump and adds the found tags.

        Args:
            text (bytes)
        """

        for line in text.split(b"\n"):
            if not line:
                continue
            parts = line.split(b"=")
            key = decode(parts[0])
            val = b"=".join(parts[1:])
            if key == "~format":
                pass
            elif key in FILESYSTEM_TAGS:
                self.add(key, bytes2fsn(val, "utf-8"))
            elif key.startswith("~#"):
                try:
                    self.add(key, int(val))
                except ValueError:
                    try:
                        self.add(key, float(val))
                    except ValueError:
                        pass
            else:
                self.add(key, decode(val))

    def change(self, key, old_value, new_value):
        """Change 'old_value' to 'new_value' for the given metadata key.
        If the old value is not found, set the key to the new value."""
        try:
            parts = self.list(key)
            try:
                parts[parts.index(old_value)] = new_value
            except ValueError:
                self[key] = new_value
            else:
                self[key] = "\n".join(parts)
        except KeyError:
            self[key] = new_value

    def add(self, key, value):
        """Add a value for the given metadata key."""
        if key not in self:
            self[key] = value
        else:
            self[key] += "\n" + value

    def remove(self, key, value=None):
        """Remove a value from the given key.

        If value is None remove all values for that key, if it exists.
        If the key or value is not found, do nothing.
        """

        if key not in self:
            return
        if value is None or self[key] == value:
            del self[key]
        else:
            try:
                parts = self.list(key)
                parts.remove(value)
                self[key] = "\n".join(parts)
            except ValueError:
                pass

    def replay_gain(self, profiles, pre_amp_gain=0, fallback_gain=0):
        """Return the computed Replay Gain scale factor.

        profiles is a list of Replay Gain profile names ('album',
        'track') to try before giving up. The special profile name
        'none' will cause no scaling to occur. pre_amp_gain will be
        applied before checking for clipping. fallback_gain will be
        used when the song does not have replaygain information.
        """
        for profile in profiles:
            if profile == "none":
                return 1.0
            try:
                db = float(self[f"replaygain_{profile}_gain"].split()[0])
                peak = float(self.get(f"replaygain_{profile}_peak", 1))
            except (KeyError, ValueError, IndexError):
                continue
            else:
                db += pre_amp_gain
                try:
                    scale = 10.0 ** (db / 20)
                except OverflowError:
                    scale = 1.0 / peak
                else:
                    if scale * peak > 1:
                        scale = 1.0 / peak  # don't clip
                return min(15, scale)
        else:
            try:
                scale = 10.0 ** ((fallback_gain + pre_amp_gain) / 20)
            except OverflowError:
                scale = 1.0
            else:
                scale = min(scale, 1.0)
            return min(15, scale)

    def write(self):
        """Write metadata back to the file.

        Raises:
            AudioFileError: in case writing fails
        """

        raise NotImplementedError

    @property
    def bookmarks(self):
        """Parse and return song position bookmarks, or set them.
        Accessing this returns a copy, so song.bookmarks.append(...)
        will not work; you need to do

            marks = song.bookmarks
            marks.append(...)
            song.bookmarks = marks
        """

        marks = []
        invalid = []
        for line in self.list("~bookmark"):
            try:
                time, mark = line.split(" ", 1)
            except Exception:
                invalid.append((-1, line))
            else:
                try:
                    time = util.parse_time(time, False)
                except Exception:
                    invalid.append((-1, line))
                else:
                    if time >= 0:
                        marks.append((time, mark))
                    else:
                        invalid.append((-1, line))
        marks.sort()
        marks.extend(invalid)
        return marks

    @bookmarks.setter
    def bookmarks(self, marks):
        result = []
        for time_, mark in marks:
            if time_ < 0:
                raise ValueError("mark times must be positive")
            result.append(f"{util.format_time(time_)} {mark}")
        result = "\n".join(result)
        if result:
            self["~bookmark"] = result
        elif "~bookmark" in self:
            del self["~bookmark"]


# Looks like the real thing.
DUMMY_SONG = AudioFile(
    {
        "~#length": 234,
        "~filename": os.devnull,
        "artist": "The Artist",
        "album": "An Example Album",
        "title": "First Track",
        "tracknumber": 1,
        "date": "2010-12-31",
    }
)



# -*- coding: utf-8 -*-
# Copyright 2016 Christoph Reiter
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import ctypes

if sys.platform == 'win32':
    from ctypes import WinDLL, CDLL, wintypes

    shell32 = WinDLL("shell32")
    kernel32 = WinDLL("kernel32")
    shlwapi = WinDLL("shlwapi")
    msvcrt = CDLL("msvcrt")

    GetCommandLineW = kernel32.GetCommandLineW
    GetCommandLineW.argtypes = []
    GetCommandLineW.restype = wintypes.LPCWSTR

    CommandLineToArgvW = shell32.CommandLineToArgvW
    CommandLineToArgvW.argtypes = [
        wintypes.LPCWSTR, ctypes.POINTER(ctypes.c_int)]
    CommandLineToArgvW.restype = ctypes.POINTER(wintypes.LPWSTR)

    LocalFree = kernel32.LocalFree
    LocalFree.argtypes = [wintypes.HLOCAL]
    LocalFree.restype = wintypes.HLOCAL

    # https://msdn.microsoft.com/en-us/library/windows/desktop/aa383751.aspx
    LPCTSTR = ctypes.c_wchar_p
    LPWSTR = wintypes.LPWSTR
    LPCWSTR = ctypes.c_wchar_p
    LPTSTR = LPWSTR
    PCWSTR = ctypes.c_wchar_p
    PCTSTR = PCWSTR
    PWSTR = ctypes.c_wchar_p
    PTSTR = PWSTR
    LPVOID = wintypes.LPVOID
    WCHAR = wintypes.WCHAR
    LPSTR = ctypes.c_char_p

    BOOL = wintypes.BOOL
    LPBOOL = ctypes.POINTER(BOOL)
    UINT = wintypes.UINT
    WORD = wintypes.WORD
    DWORD = wintypes.DWORD
    SHORT = wintypes.SHORT
    HANDLE = wintypes.HANDLE
    ULONG = wintypes.ULONG
    LPCSTR = wintypes.LPCSTR

    STD_INPUT_HANDLE = DWORD(-10)
    STD_OUTPUT_HANDLE = DWORD(-11)
    STD_ERROR_HANDLE = DWORD(-12)

    INVALID_HANDLE_VALUE = wintypes.HANDLE(-1).value

    INTERNET_MAX_SCHEME_LENGTH = 32
    INTERNET_MAX_PATH_LENGTH = 2048
    INTERNET_MAX_URL_LENGTH = (
        INTERNET_MAX_SCHEME_LENGTH + len("://") + INTERNET_MAX_PATH_LENGTH)

    FOREGROUND_BLUE = 0x0001
    FOREGROUND_GREEN = 0x0002
    FOREGROUND_RED = 0x0004
    FOREGROUND_INTENSITY = 0x0008

    BACKGROUND_BLUE = 0x0010
    BACKGROUND_GREEN = 0x0020
    BACKGROUND_RED = 0x0040
    BACKGROUND_INTENSITY = 0x0080

    COMMON_LVB_REVERSE_VIDEO = 0x4000
    COMMON_LVB_UNDERSCORE = 0x8000

    UrlCreateFromPathW = shlwapi.UrlCreateFromPathW
    UrlCreateFromPathW.argtypes = [
        PCTSTR, PTSTR, ctypes.POINTER(DWORD), DWORD]
    UrlCreateFromPathW.restype = ctypes.HRESULT

    SetEnvironmentVariableW = kernel32.SetEnvironmentVariableW
    SetEnvironmentVariableW.argtypes = [LPCTSTR, LPCTSTR]
    SetEnvironmentVariableW.restype = wintypes.BOOL

    GetEnvironmentVariableW = kernel32.GetEnvironmentVariableW
    GetEnvironmentVariableW.argtypes = [LPCTSTR, LPTSTR, DWORD]
    GetEnvironmentVariableW.restype = DWORD

    GetEnvironmentStringsW = kernel32.GetEnvironmentStringsW
    GetEnvironmentStringsW.argtypes = []
    GetEnvironmentStringsW.restype = ctypes.c_void_p

    FreeEnvironmentStringsW = kernel32.FreeEnvironmentStringsW
    FreeEnvironmentStringsW.argtypes = [ctypes.c_void_p]
    FreeEnvironmentStringsW.restype = ctypes.c_bool

    GetStdHandle = kernel32.GetStdHandle
    GetStdHandle.argtypes = [DWORD]
    GetStdHandle.restype = HANDLE

    class COORD(ctypes.Structure):

        _fields_ = [
            ("X", SHORT),
            ("Y", SHORT),
        ]

    class SMALL_RECT(ctypes.Structure):

        _fields_ = [
            ("Left", SHORT),
            ("Top", SHORT),
            ("Right", SHORT),
            ("Bottom", SHORT),
        ]

    class CONSOLE_SCREEN_BUFFER_INFO(ctypes.Structure):

        _fields_ = [
            ("dwSize", COORD),
            ("dwCursorPosition", COORD),
            ("wAttributes", WORD),
            ("srWindow", SMALL_RECT),
            ("dwMaximumWindowSize", COORD),
        ]

    GetConsoleScreenBufferInfo = kernel32.GetConsoleScreenBufferInfo
    GetConsoleScreenBufferInfo.argtypes = [
        HANDLE, ctypes.POINTER(CONSOLE_SCREEN_BUFFER_INFO)]
    GetConsoleScreenBufferInfo.restype = BOOL

    GetConsoleOutputCP = kernel32.GetConsoleOutputCP
    GetConsoleOutputCP.argtypes = []
    GetConsoleOutputCP.restype = UINT

    SetConsoleOutputCP = kernel32.SetConsoleOutputCP
    SetConsoleOutputCP.argtypes = [UINT]
    SetConsoleOutputCP.restype = BOOL

    GetConsoleCP = kernel32.GetConsoleCP
    GetConsoleCP.argtypes = []
    GetConsoleCP.restype = UINT

    SetConsoleCP = kernel32.SetConsoleCP
    SetConsoleCP.argtypes = [UINT]
    SetConsoleCP.restype = BOOL

    SetConsoleTextAttribute = kernel32.SetConsoleTextAttribute
    SetConsoleTextAttribute.argtypes = [HANDLE, WORD]
    SetConsoleTextAttribute.restype = BOOL

    SetConsoleCursorPosition = kernel32.SetConsoleCursorPosition
    SetConsoleCursorPosition.argtypes = [HANDLE, COORD]
    SetConsoleCursorPosition.restype = BOOL

    ReadConsoleW = kernel32.ReadConsoleW
    ReadConsoleW.argtypes = [
        HANDLE, LPVOID, DWORD, ctypes.POINTER(DWORD), LPVOID]
    ReadConsoleW.restype = BOOL

    MultiByteToWideChar = kernel32.MultiByteToWideChar
    MultiByteToWideChar.argtypes = [
        UINT, DWORD, LPCSTR, ctypes.c_int, LPWSTR, ctypes.c_int]
    MultiByteToWideChar.restype = ctypes.c_int

    WideCharToMultiByte = kernel32.WideCharToMultiByte
    WideCharToMultiByte.argtypes = [
        UINT, DWORD, LPCWSTR, ctypes.c_int, LPSTR, ctypes.c_int,
        LPCSTR, LPBOOL]
    WideCharToMultiByte.restype = ctypes.c_int

    MoveFileW = kernel32.MoveFileW
    MoveFileW.argtypes = [LPCTSTR, LPCTSTR]
    MoveFileW.restype = BOOL

    GetFileInformationByHandleEx = None
    if hasattr(kernel32, "GetFileInformationByHandleEx"):
        GetFileInformationByHandleEx = kernel32.GetFileInformationByHandleEx
        GetFileInformationByHandleEx.argtypes = [
            HANDLE, ctypes.c_int, ctypes.c_void_p, DWORD]
        GetFileInformationByHandleEx.restype = BOOL
    else:
        # Windows XP
        pass

    MAX_PATH = 260
    FileNameInfo = 2

    class FILE_NAME_INFO(ctypes.Structure):
        _fields_ = [
            ("FileNameLength", DWORD),
            ("FileName", WCHAR),
        ]

    _get_osfhandle = msvcrt._get_osfhandle
    _get_osfhandle.argtypes = [ctypes.c_int]
    _get_osfhandle.restype = HANDLE

    GetFileType = kernel32.GetFileType
    GetFileType.argtypes = [HANDLE]
    GetFileType.restype = DWORD

    FILE_TYPE_PIPE = 0x0003



# Copyright 2012-2023 Nick Boultbee
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from gi.repository import Gtk

import os

from quodlibet.util.songwrapper import SongWrapper

from quodlibet.qltk.songsmenu import confirm_multi_song_invoke

import quodlibet
from quodlibet import _
from quodlibet import qltk
from quodlibet import util
from quodlibet.pattern import Pattern
from quodlibet.plugins import PluginConfigMixin
from quodlibet.plugins.playlist import PlaylistPlugin
from quodlibet.plugins.songsmenu import SongsMenuPlugin
from quodlibet.qltk.data_editors import JSONBasedEditor
from quodlibet.qltk.x import SeparatorMenuItem
from quodlibet.qltk import ErrorMessage, Icons
from quodlibet.qltk.getstring import GetStringDialog
from quodlibet.util.dprint import print_w, print_d, print_e
from quodlibet.util.json_data import JSONObject, JSONObjectDict
from quodlibet.util import connect_obj, print_exc

Field = JSONObject.Field


class Command(JSONObject):
    """
    Wraps an arbitrary shell command and its argument pattern.
    Serialises as JSON for some editability
    """

    NAME = _("Command")

    FIELDS = {
        "name": Field(_("name"), _("The name of this command")),
        "command": Field(_("command"), _("The shell command syntax to run")),
        "parameter": Field(
            _("parameter"),
            _(
                "If specified, a parameter whose occurrences in "
                "the command will be substituted with a "
                "user-supplied value, e.g. by using 'PARAM' "
                "all instances of '{PARAM}' in your command will "
                "have the value prompted for when run"
            ),
        ),
        "pattern": Field(
            _("pattern"),
            _(
                "The QL pattern, e.g. <~filename>, to use to "
                "compute a value for the command. For playlists, "
                "this also supports virtual tags <~playlistname> "
                "and <~#playlistindex>."
            ),
        ),
        "unique": Field(
            _("unique"),
            _("If set, this will remove duplicate computed values of the pattern"),
        ),
        "max_args": Field(
            _("max args"),
            _(
                "The maximum number of argument to pass to the "
                "command at one time (like xargs)"
            ),
        ),
        "reverse": Field(_("reverse"), _("If set, the argument list will be reversed")),
    }

    def __init__(
        self,
        name=None,
        command=None,
        pattern="<~filename>",
        unique=False,
        parameter=None,
        max_args=10000,
        reverse=False,
        warn_threshold=50,
    ):
        JSONObject.__init__(self, name)
        self.command = str(command or "")
        self.pattern = str(pattern)
        self.unique = bool(unique)
        self.max_args = max_args
        self.reverse = bool(reverse)
        self.parameter = str(parameter or "")
        self.__pat = Pattern(self.pattern)
        self.warn_threshold = warn_threshold

    def run(self, songs, playlist_name=None):
        """
        Runs this command on `songs`,
        splitting into multiple calls if necessary.
        `playlist_name` if populated contains the Playlist's name.
        """
        args = []
        template_vars = {}
        if self.parameter:
            value = GetStringDialog(
                None, _("Input value"), _("Value for %s?") % self.parameter
            ).run()
            template_vars[self.parameter] = value
        if playlist_name:
            print_d(f"Playlist command for {playlist_name}")
            template_vars["PLAYLIST"] = playlist_name

        actual_command = self.command.format(**template_vars)
        print_d(f"Actual command={actual_command}")
        for i, song in enumerate(songs):
            wrapped = SongWrapper(song)
            if playlist_name:
                wrapped["~playlistname"] = playlist_name
                wrapped["~playlistindex"] = str(i + 1)
                wrapped["~#playlistindex"] = i + 1
            arg = str(self.__pat.format(wrapped))
            if not arg:
                print_w(
                    f'Couldn\'t build shell command using "{self.pattern}".'
                    "Check your pattern?"
                )
                break
            if not self.unique:
                args.append(arg)
            elif arg not in args:
                args.append(arg)
        max = int(self.max_args or 10000)
        com_words = actual_command.split(" ")

        if self.reverse:
            args.reverse()

        while args:
            print_d(
                "Running %s with %d substituted arg(s) (of %d%s total)..."
                % (
                    actual_command,
                    min(max, len(args)),
                    len(args),
                    " unique" if self.unique else "",
                )
            )
            util.spawn(com_words + args[:max])
            args = args[max:]

    @property
    def playlists_only(self):
        return "~playlistname" in self.pattern or "playlistindex" in self.pattern

    def __str__(self):
        return "{command} {pattern}".format(**dict(self.data))


class CustomCommands(PlaylistPlugin, SongsMenuPlugin, PluginConfigMixin):
    PLUGIN_ICON = Icons.APPLICATION_UTILITIES
    PLUGIN_ID = "CustomCommands"
    PLUGIN_NAME = _("Custom Commands")
    PLUGIN_DESC = _(
        "Runs custom commands (in batches if required) on songs "
        "using any of their tags."
    )

    # Here are some starters...
    DEFAULT_COMS = [
        Command("Compress files", "file-roller -d"),
        Command(
            "Browse folders (Thunar)",
            "thunar",
            "<~dirname>",
            unique=True,
            max_args=50,
            warn_threshold=20,
        ),
        Command("Burn using K3b", "k3b --audiocd"),
        Command("Burn using Brasero", "brasero --audio"),
        Command("Burn using Xfburn", "xfburn --audio-composition", reverse=True),
        Command(
            "Flash notification",
            command="notify-send"
            " -t 2000"
            " -i "
            "/usr/share/icons/hicolor/scalable/apps/"
            "io.github.quodlibet.QuodLibet.svg",
            pattern='<~rating> "<title><version| (<version>)>"'
            "<~people| by <~people>>"
            "<album|, from <album><discnumber| : disk <discnumber>>"
            "<~length| (<~length>)>",
            max_args=1,
            warn_threshold=10,
        ),
        Command(
            "Output playlist to stdout",
            command="echo -e",
            pattern="<~playlistname>: <~playlistindex>.  <~artist~title>\\\\n",
            warn_threshold=20,
        ),
        Command("Fix MP3 VBR with mp3val", "mp3val -f", unique=True, max_args=1),
        Command(
            "Record Stream",
            command="x-terminal-emulator -e wget -P $HOME",
            pattern="<~filename>",
            max_args=1,
        ),
    ]

    COMS_FILE = os.path.join(quodlibet.get_user_dir(), "lists", "customcommands.json")

    _commands = None
    """Commands known to the class"""

    def __set_pat(self, name):
        self.com_index = name

    def get_data(self, key):
        """Gets the pattern for a given key"""
        try:
            return self.all_commands()[key]
        except (KeyError, TypeError):
            print_d(f"Invalid key {key}")
            return None

    @classmethod
    def edit_patterns(cls, button):
        win = JSONBasedEditor(
            Command,
            cls.all_commands(),
            filename=cls.COMS_FILE,
            title=_("Edit Custom Commands"),
        )
        # Cache busting
        cls._commands = None
        win.show()

    @classmethod
    def PluginPreferences(cls, parent):
        hb = Gtk.HBox(spacing=3)
        hb.set_border_width(0)

        button = qltk.Button(_("Edit Custom Commands") + "…", Icons.EDIT)
        button.connect("clicked", cls.edit_patterns)
        hb.pack_start(button, True, True, 0)
        hb.show_all()
        return hb

    @classmethod
    def all_commands(cls):
        if cls._commands is None:
            cls._commands = cls._get_saved_commands()
        return cls._commands

    @classmethod
    def _get_saved_commands(cls):
        filename = cls.COMS_FILE
        print_d(f"Loading saved commands from '{filename}'...")
        coms = None
        try:
            with open(filename, encoding="utf-8") as f:
                coms = JSONObjectDict.from_json(Command, f.read())
        except (OSError, ValueError) as e:
            print_w(f"Couldn't parse saved commands ({e})")

        # Failing all else...
        if not coms:
            print_d(f"No commands found in {filename}. Using defaults.")
            coms = {c.name: c for c in cls.DEFAULT_COMS}
        print_d(f"Loaded commands: {coms.keys()}")
        return coms

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        pl_mode = hasattr(self, "_playlists") and bool(len(self._playlists))
        self.com_index = None
        self.unique_only = False
        submenu = Gtk.Menu()
        for name, c in self.all_commands().items():
            item = Gtk.MenuItem(label=name)
            connect_obj(item, "activate", self.__set_pat, name)
            if pl_mode and not c.playlists_only:
                continue
            item.set_sensitive(c.playlists_only == pl_mode)
            submenu.append(item)

        self.add_edit_item(submenu)
        if submenu.get_children():
            self.set_submenu(submenu)
        else:
            self.set_sensitive(False)

    @classmethod
    def add_edit_item(cls, submenu):
        config = Gtk.MenuItem(label=_("Edit Custom Commands") + "…")
        connect_obj(config, "activate", cls.edit_patterns, config)
        config.set_sensitive(not JSONBasedEditor.is_not_unique())
        submenu.append(SeparatorMenuItem())
        submenu.append(config)

    def plugin_songs(self, songs):
        self._handle_songs(songs)

    def plugin_playlist(self, playlist):
        print_d(f"Running playlist plugin for {playlist}")
        return self._handle_songs(playlist.songs, playlist)

    def _handle_songs(self, songs, playlist=None):
        # Check this is a launch, not a configure
        if self.com_index:
            com = self.get_data(self.com_index)
            if len(songs) > com.warn_threshold:
                if not confirm_multi_song_invoke(self, com.name, len(songs)):
                    print_d("User decided not to run on %d songs" % len(songs))
                    return
            print_d("Running %s on %d song(s)" % (com, len(songs)))
            try:
                com.run(songs, playlist and playlist.name)
            except Exception as err:
                print_e(f"Couldn't run command {com.name}: {type(err)} {err} at:")
                print_exc()
                ErrorMessage(
                    self.plugin_window,
                    _("Unable to run custom command %s") % util.escape(self.com_index),
                    util.escape(str(err)),
                ).run()



# Copyright 2014 Christoph Reiter
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


import collections
import threading


class Logs:
    """Thread safe log store"""

    MAX_LOG_SIZE_DEFAULT = 500

    def __init__(self, max_log_size=MAX_LOG_SIZE_DEFAULT):
        self._iter_lock = threading.Lock()
        self._log = collections.deque(maxlen=max_log_size)

    def _save_iter(self):
        # only pop/append/len are threadsafe, implement iter with them
        with self._iter_lock:
            temp = collections.deque()
            for _i in range(len(self._log)):
                item = self._log.popleft()
                yield item
                temp.append(item)

            while temp:
                self._log.appendleft(temp.pop())

    def log(self, string, category=None):
        """Log str/unicode.

        Thread safe.
        """

        self._log.append((category, string))

    def clear(self):
        """Remove all entries.

        Thread safe.
        """

        with self._iter_lock:
            for _i in range(len(self._log)):
                self._log.popleft()

    def get_content(self, category=None, limit=None):
        """Get a list of unicode strings for the specified category.
        Oldest entry first. Passing no category will return all content.
        If `limit` is specified, the last `limit` items will be returned.

        Thread safe.
        """

        content = []
        for cat, string in self._save_iter():
            if category is None or category == cat:
                if isinstance(string, bytes):
                    string = string.decode("utf-8", "replace")
                content.append(string)

        if limit is not None:
            assert limit > 0
            return content[-limit:]
        return content


_logs = Logs()
log = _logs.log
get_content = _logs.get_content
